"""

RenderPipeline

Copyright (c) 2014-2016 tobspr <tobias.springer1@gmail.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

"""

from panda3d.core import Vec2, Vec3, Vec4
from panda3d.core import TextNode as TextNodeImpl

from metadrive.render_pipeline.rpcore.globals import Globals
from metadrive.render_pipeline.rpcore.rpobject import RPObject
from metadrive.render_pipeline.rpcore.loader import RPLoader


class TextNode(RPObject):
    """ Interface for the Panda3D TextNode. """
    def __init__(
        self,
        font="/$$rp/data/font/Roboto-Bold.ttf",
        pixel_size=16,
        align="left",
        pos=Vec2(0),
        color=Vec3(1),
        parent=None
    ):
        """ Constructs a new text node, forwaring the parameters to the internal
        panda3d implementation """
        RPObject.__init__(self)
        self._node = TextNodeImpl('FTN')
        self._node.set_text("")
        self._node.set_align(getattr(TextNodeImpl, "A_" + align))
        self._node.set_text_color(color.x, color.y, color.z, 1)

        if parent is None:
            parent = Globals.base.aspect2d

        self._nodepath = parent.attach_new_node(self._node)
        self._nodepath.set_pos(pos.x, 0, pos.y)

        font = RPLoader.load_font(font)
        # font.set_outline(Vec4(0, 0, 0, 0.78), 1.6, 0.37)
        font.set_outline(Vec4(0, 0, 0, 1), 1.6, 0.37)
        font.set_scale_factor(1.0)
        font.set_texture_margin(int(pixel_size / 4.0 * 2.0))
        font.set_bg(Vec4(0, 0, 0, 0))
        self._node.set_font(font)
        self.set_pixel_size(pixel_size)

    @property
    def text(self):
        """ Returns the current text """
        return self._node.get_text()

    @text.setter
    def text(self, text):
        """ Sets the current text """
        self._node.set_text(text)

    @property
    def color(self):
        """ Returns the current text color """
        return self._node.get_text_color()

    @color.setter
    def color(self, val):
        """ Sets the current text color """
        self._node.set_text_color(val)

    def set_pixel_size(self, size):
        """ Sets the text size in pixels """
        self._nodepath.set_scale(size * 2.0 / float(Globals.native_resolution.y))



# Copyright (c) 2020 kitsune.ONE team.

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import bpy
import mathutils  # make sure to "import bpy" before

from metadrive.third_party.kitsunetsuki.base.vertex import uv_equals, normal_equals


class VertexMixin(object):
    def get_sharp_vertices(self, mesh):
        results = []
        if mesh.use_auto_smooth:
            for edge in mesh.edges:
                if edge.use_edge_sharp:
                    for vertex_id in edge.vertices:
                        results.append(vertex_id)

        return results

    def get_tangent_bitangent(self, mesh):
        results = {}

        for uv_name, uv_layer in mesh.uv_layers.items():
            mesh.calc_tangents(uvmap=uv_name)
            results[uv_name] = []
            for i, loop in mesh.loops.items():
                results[uv_name].append(
                    (
                        mathutils.Vector(loop.tangent),
                        mathutils.Vector(loop.bitangent),
                        loop.bitangent_sign,
                    )
                )
            mesh.free_tangents()

        return results

    def can_share_vertex(self, mesh, vertex, loop_id, uv, normal):
        if not mesh.uv_layers:
            return True

        if not mesh.uv_layers.active:
            return True

        uv_loop = mesh.uv_layers.active.data[loop_id]
        # if uv_equals(uv_loop.uv.to_2d(), uv) and normal_equals(vertex.normal, normal):
        if uv_equals(uv_loop.uv.to_2d(), uv) and normal_equals(mesh.loops[loop_id].normal, normal):
            return True

        return False



from metadrive.envs.metadrive_env import MetaDriveEnv
from metadrive.utils import recursive_equal, setup_logger


def test_gen_map_alignment():
    """
    For visualization check
    """
    env_num = 2
    step = 1000
    generate_config = {"num_scenarios": env_num, "start_seed": 0, "use_render": False}

    setup_logger(debug=True)
    try:
        env = MetaDriveEnv(generate_config)
        env.reset()
        data_1 = env.engine.map_manager.dump_all_maps(file_name="test_10maps.pickle")
        env.close()
        env = None

        env = MetaDriveEnv(generate_config)
        for i in range(env_num):
            env.reset(seed=i)
        data_2 = env.engine.map_manager.dump_all_maps(file_name="test_10maps.pickle")
        recursive_equal(data_1.copy(), data_2.copy(), need_assert=True)

    finally:
        env.close()


if __name__ == "__main__":
    test_gen_map_alignment()



from tqdm import tqdm

from metadrive.engine.asset_loader import AssetLoader
from metadrive.envs.real_data_envs.nuscenes_env import NuScenesEnv
from metadrive.policy.replay_policy import ReplayEgoCarPolicy


def _test_level(level=1, render=False):
    env = NuScenesEnv(
        {
            "use_render": render,
            "agent_policy": ReplayEgoCarPolicy,
            "sequential_seed": True,
            "reactive_traffic": False,
            "window_size": (1600, 900),
            "num_scenarios": 10,
            "horizon": 1000,
            "curriculum_level": level,
            "no_static_vehicles": True,
            "data_directory": AssetLoader.file_path("nuscenes", return_raw_style=False),
        }
    )
    try:
        scenario_id = set()
        for i in tqdm(range(10), desc=str(level)):
            env.reset(seed=i)
            for i in range(10):
                o, r, d, _, _ = env.step([0, 0])
                if d:
                    break

            scenario_id.add(env.engine.data_manager.current_scenario_summary["id"])
        assert len(scenario_id) == (env.engine.current_level + 1) * int(10 / level)
    finally:
        env.close()


def test_curriculum_seed():
    _test_level(level=5)
    _test_level(level=1)
    _test_level(level=2)
    # _test_level(level=3)


def test_curriculum_up_1_level(render=False, level=5):
    env = NuScenesEnv(
        {
            "use_render": render,
            "agent_policy": ReplayEgoCarPolicy,
            "sequential_seed": True,
            "reactive_traffic": False,
            "window_size": (1600, 900),
            "num_scenarios": 10,
            "episodes_to_evaluate_curriculum": 2,
            "horizon": 1000,
            "curriculum_level": level,
            "no_static_vehicles": True,
            "data_directory": AssetLoader.file_path("nuscenes", return_raw_style=False),
        }
    )
    try:
        scenario_id = []
        for i in tqdm(range(10), desc=str(level)):
            env.reset(seed=i)
            for i in range(10):
                o, r, d, _, _ = env.step([0, 0])
            scenario_id.append(env.engine.data_manager.current_scenario_summary["id"])
        assert len(set(scenario_id)) == 4
        ids = [env.engine.data_manager.summary_dict[f]["id"] for f in env.engine.data_manager.summary_lookup]
        assert set(scenario_id) == set(ids[:4])
    finally:
        env.close()


def test_curriculum_level_up(render=False):
    env = NuScenesEnv(
        {
            "use_render": render,
            "agent_policy": ReplayEgoCarPolicy,
            "sequential_seed": True,
            "reactive_traffic": False,
            "window_size": (1600, 900),
            "num_scenarios": 10,
            "episodes_to_evaluate_curriculum": 5,
            "target_success_rate": 1.,
            "horizon": 1000,
            "curriculum_level": 2,
            "no_static_vehicles": True,
            "data_directory": AssetLoader.file_path("nuscenes", return_raw_style=False),
        }
    )
    try:
        scenario_id = []
        for i in tqdm(range(20), desc=str(2)):
            env.reset()
            for i in range(250):
                o, r, d, _, _ = env.step([0, 0])
            scenario_id.append(env.engine.data_manager.current_scenario_summary["id"])
        assert len(set(scenario_id)) == 10
        ids = [env.engine.data_manager.summary_dict[f]["id"] for f in env.engine.data_manager.summary_lookup]
        assert scenario_id[:10] == ids
        assert scenario_id[-5:] == ids[-5:] == scenario_id[-10:-5]
    finally:
        env.close()


def _worker_env(render, worker_index, level_up=False):
    assert worker_index in [0, 1]
    level = 2
    env = NuScenesEnv(
        {
            "use_render": render,
            "agent_policy": ReplayEgoCarPolicy,
            "sequential_seed": True,
            "reactive_traffic": False,
            "window_size": (1600, 900),
            "num_scenarios": 8,
            "episodes_to_evaluate_curriculum": 4,
            "curriculum_level": level,
            "no_static_vehicles": True,
            "data_directory": AssetLoader.file_path("nuscenes", return_raw_style=False),
            "worker_index": worker_index,
            "num_workers": 2,
        }
    )
    try:
        env.reset()
        if level_up:
            env.engine.curriculum_manager._level_up()
            env.reset()
        scenario_id = []
        for i in range(20):
            env.reset()
            for i in range(10):
                o, r, d, _, _ = env.step([0, 0])
            scenario_id.append(env.engine.data_manager.current_scenario_summary["id"])
            print(env.current_seed)
        all_scenario = [
            env.engine.data_manager.summary_dict[f]["id"] for f in env.engine.data_manager.summary_lookup[:8]
        ]
        assert len(set(scenario_id)) == 2
        assert env.engine.data_manager.data_coverage == 0.75 if level_up else 0.5
    finally:
        env.close()
    return scenario_id, all_scenario


def test_curriculum_multi_worker(render=False):
    # 1
    all_scenario_id = []
    all_scenario_id.extend(_worker_env(render, 1, level_up=True)[0])
    set_1, all_scenario = _worker_env(render, 0, level_up=True)
    all_scenario_id.extend(set_1)

    ids = all_scenario
    assert set(all_scenario_id) == set(ids[-4:])
    # 2
    all_scenario_id = []
    set_1, all_scenario = _worker_env(render, 0)
    all_scenario_id.extend(set_1)
    all_scenario_id.extend(_worker_env(render, 1)[0])

    ids = all_scenario
    assert set(all_scenario_id) == set(ids[:4])


def level_up_worker(render, worker_index):
    env = NuScenesEnv(
        {
            "use_render": render,
            "agent_policy": ReplayEgoCarPolicy,
            "sequential_seed": True,
            "reactive_traffic": False,
            "window_size": (1600, 900),
            "num_scenarios": 8,
            "episodes_to_evaluate_curriculum": 2,
            "horizon": 1000,
            "curriculum_level": 4,
            "no_static_vehicles": True,
            "num_workers": 2,
            "worker_index": worker_index,
            "data_directory": AssetLoader.file_path("nuscenes", return_raw_style=False),
        }
    )
    try:
        scenario_id = []
        for i in tqdm(range(20), desc=str(2)):
            env.reset()
            for i in range(250):
                o, r, d, _, _ = env.step([0, 0])
            scenario_id.append(env.engine.data_manager.current_scenario_summary["id"])
        assert len(set(scenario_id)) == 4
        ids = [env.engine.data_manager.summary_dict[f]["id"] for f in env.engine.data_manager.summary_lookup[:8]]
        assert scenario_id[:4] == ids[worker_index::2]
        assert scenario_id[-5:] == [ids[-2] if worker_index == 0 else ids[-1]] * 5 == scenario_id[-10:-5]
    finally:
        env.close()
    return scenario_id, ids


def test_curriculum_worker_level_up(render=False):
    scenario_id_1, all_id = level_up_worker(render, 0)
    scenario_id_2, all_id = level_up_worker(render, 1)
    assert scenario_id_1[:4] + scenario_id_2[:4] == all_id[::2] + all_id[1::2]


def test_start_seed_not_0(render=False, worker_index=0):
    env = NuScenesEnv(
        {
            "use_render": render,
            "agent_policy": ReplayEgoCarPolicy,
            "sequential_seed": True,
            "reactive_traffic": False,
            "window_size": (1600, 900),
            "num_scenarios": 8,
            "start_scenario_index": 2,
            "episodes_to_evaluate_curriculum": 2,
            "horizon": 1000,
            "curriculum_level": 4,
            "no_static_vehicles": True,
            "num_workers": 2,
            "worker_index": worker_index,
            "data_directory": AssetLoader.file_path("nuscenes", return_raw_style=False),
        }
    )
    try:
        scenario_id = []
        for i in tqdm(range(20), desc=str(2)):
            env.reset()
            for i in range(250):
                o, r, d, _, _ = env.step([0, 0])
            scenario_id.append(env.engine.data_manager.current_scenario_summary["id"])
        all_scenarios = sorted(list(env.engine.data_manager.summary_dict.keys()))[2:]
        summary_lookup = env.engine.data_manager.summary_lookup[2:]
        assert set(all_scenarios) == set(summary_lookup)
        assert len(set(scenario_id)) == 4
        ids = [env.engine.data_manager.summary_dict[f]["id"] for f in summary_lookup]
        assert scenario_id[:4] == ids[worker_index::2]
        assert scenario_id[-5:] == [ids[-2] if worker_index == 0 else ids[-1]] * 5 == scenario_id[-10:-5]
    finally:
        env.close()
    return scenario_id, ids


def test_start_seed_1_9(render=False, worker_index=0):
    env = NuScenesEnv(
        {
            "use_render": render,
            "agent_policy": ReplayEgoCarPolicy,
            "sequential_seed": True,
            "reactive_traffic": False,
            "window_size": (1600, 900),
            "num_scenarios": 8,
            "start_scenario_index": 1,
            "episodes_to_evaluate_curriculum": 2,
            "horizon": 1000,
            "curriculum_level": 4,
            "no_static_vehicles": True,
            "num_workers": 2,
            "worker_index": worker_index,
            "data_directory": AssetLoader.file_path("nuscenes", return_raw_style=False),
        }
    )
    try:
        scenario_id = []
        for i in tqdm(range(20), desc=str(2)):
            env.reset()
            for i in range(250):
                o, r, d, _, _ = env.step([0, 0])
            scenario_id.append(env.engine.data_manager.current_scenario_summary["id"])
        all_scenarios = sorted(list(env.engine.data_manager.summary_dict.keys()))[1:9]
        summary_lookup = env.engine.data_manager.summary_lookup[1:9]
        assert set(all_scenarios) == set(summary_lookup)
        assert len(set(scenario_id)) == 4
        ids = [env.engine.data_manager.summary_dict[f]["id"] for f in summary_lookup]
        assert scenario_id[:4] == ids[worker_index::2]
        assert scenario_id[-5:] == [ids[-2] if worker_index == 0 else ids[-1]] * 5 == scenario_id[-10:-5]
    finally:
        env.close()
    return scenario_id, ids


if __name__ == '__main__':
    test_curriculum_multi_worker()
    # test_curriculum_seed()
    # test_curriculum_level_up()
    # test_curriculum_worker_level_up()
    # test_start_seed_not_0()
    # test_start_seed_1_9()



from metadrive.component.vehicle.base_vehicle import BaseVehicle
from metadrive.component.vehicle.vehicle_type import DefaultVehicle
from metadrive.constants import MetaDriveType
from metadrive.constants import DEFAULT_AGENT
from metadrive.envs.metadrive_env import MetaDriveEnv
from metadrive.utils import setup_logger


def test_original_lidar(render=False):
    setup_logger(debug=True)
    env = MetaDriveEnv(
        {
            "use_render": render,
            "manual_control": render,
            "num_scenarios": 1,
            "traffic_density": 0.3,
            "vehicle_config": {
                "show_lidar": True,
                "side_detector": dict(num_lasers=2, distance=50),
                "lane_line_detector": dict(num_lasers=2, distance=50),
            },
            "_disable_detector_mask": True,
            "map": "XXX"
        }
    )
    try:
        env.reset()
        v_config = env.config["vehicle_config"]
        v_config["spawn_longitude"] = 0
        v_config["spawn_lateral"] = 7.5
        another_v = DefaultVehicle(v_config, random_seed=0)
        another_v.reset()
        objs = env.vehicle.side_detector.perceive(env.vehicle, env.vehicle.engine.physics_world.static_world
                                                  ).detected_objects + env.vehicle.lane_line_detector.perceive(
                                                      env.vehicle, env.vehicle.engine.physics_world.static_world
                                                  ).detected_objects
        yellow = 0
        for obj in objs:
            if obj.getNode().getName() == MetaDriveType.LINE_SOLID_SINGLE_YELLOW:
                yellow += 1
        assert yellow == 2, "side detector and lane detector broken"
        detect_traffic_vehicle = False
        detect_base_vehicle = False
        for i in range(1, 1000):
            o, r, tm, tc, info = env.step([0, 1])
            if len(env.vehicle.lidar.get_surrounding_vehicles(env.observations[DEFAULT_AGENT].detected_objects)) > 2:
                detect_traffic_vehicle = True
            for hit in env.observations[DEFAULT_AGENT].detected_objects:
                if isinstance(hit, BaseVehicle):
                    detect_base_vehicle = True
            if tm or tc:
                break
        # if not (detect_traffic_vehicle and detect_base_vehicle):
        #     print("Lidar detection failed")
        assert detect_traffic_vehicle and detect_base_vehicle, "Lidar detection failed"
    finally:
        env.close()


def test_lidar_with_mask(render=False):
    setup_logger(debug=True)
    env = MetaDriveEnv(
        {
            "use_render": render,
            "manual_control": render,
            "num_scenarios": 1,
            "traffic_density": 0.3,
            "vehicle_config": {
                "show_lidar": True,
                "side_detector": dict(num_lasers=2, distance=50),
                "lane_line_detector": dict(num_lasers=2, distance=50),
            },
            "_disable_detector_mask": False,
            "map": "XXX"
        }
    )
    try:
        env.reset()
        v_config = env.config["vehicle_config"]
        v_config["spawn_longitude"] = 0
        v_config["spawn_lateral"] = 7.5
        another_v = DefaultVehicle(v_config, random_seed=0)
        another_v.reset()
        # for test
        objs = env.vehicle.side_detector.perceive(env.vehicle, env.vehicle.engine.physics_world.static_world
                                                  ).detected_objects + env.vehicle.lane_line_detector.perceive(
                                                      env.vehicle, env.vehicle.engine.physics_world.static_world
                                                  ).detected_objects
        yellow = 0
        for obj in objs:
            if obj.getNode().getName() == MetaDriveType.LINE_SOLID_SINGLE_YELLOW:
                yellow += 1
        assert yellow == 2, "side detector and lane detector broken"
        detect_traffic_vehicle = False
        detect_base_vehicle = False
        for i in range(1, 1000):
            o, r, tm, tc, info = env.step([0, 1])
            if len(env.vehicle.lidar.get_surrounding_vehicles(env.observations[DEFAULT_AGENT].detected_objects)) > 2:
                detect_traffic_vehicle = True
            for hit in env.observations[DEFAULT_AGENT].detected_objects:
                if isinstance(hit, BaseVehicle):
                    detect_base_vehicle = True
            if tm or tc:
                break
        # if not (detect_traffic_vehicle and detect_base_vehicle):
        #     print("Lidar detection failed")
        assert detect_traffic_vehicle and detect_base_vehicle, "Lidar detection failed"
    finally:
        env.close()


if __name__ == "__main__":
    # test_lidar_with_mask(render=True)
    test_original_lidar(render=True)



import copy
import os
import pickle

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.pyplot import figure

from metadrive.component.static_object.traffic_object import TrafficCone, TrafficBarrier
from metadrive.component.traffic_light.base_traffic_light import BaseTrafficLight
from metadrive.component.traffic_participants.cyclist import Cyclist
from metadrive.component.traffic_participants.pedestrian import Pedestrian
from metadrive.component.vehicle.base_vehicle import BaseVehicle
from metadrive.constants import DATA_VERSION, DEFAULT_AGENT
from metadrive.scenario import ScenarioDescription as SD
from metadrive.scenario.scenario_description import ScenarioDescription
from metadrive.type import MetaDriveType
from metadrive.utils.math import wrap_to_pi

NP_ARRAY_DECIMAL = 3
VELOCITY_DECIMAL = 1  # velocity can not be set accurately
MIN_LENGTH_RATIO = 0.8


def draw_map(map_features, show=False):
    figure(figsize=(8, 6), dpi=500)
    for key, value in map_features.items():
        if value.get("type", None) == MetaDriveType.LANE_SURFACE_STREET:
            plt.scatter([x[0] for x in value["polyline"]], [y[1] for y in value["polyline"]], s=0.1)
        elif value.get("type", None) == "road_edge":
            plt.scatter([x[0] for x in value["polyline"]], [y[1] for y in value["polyline"]], s=0.1, c=(0, 0, 0))
        # elif value.get("type", None) == "road_line":
        #     plt.scatter([x[0] for x in value["polyline"]], [y[1] for y in value["polyline"]], s=0.5, c=(0.8,0.8,0.8))
    if show:
        plt.show()


def get_type_from_class(obj_class):
    if issubclass(obj_class, BaseVehicle) or obj_class is BaseVehicle:
        return MetaDriveType.VEHICLE
    elif issubclass(obj_class, Pedestrian) or obj_class is Pedestrian:
        return MetaDriveType.PEDESTRIAN
    elif issubclass(obj_class, Cyclist) or obj_class is Cyclist:
        return MetaDriveType.CYCLIST
    elif issubclass(obj_class, BaseTrafficLight) or obj_class is BaseTrafficLight:
        return MetaDriveType.TRAFFIC_LIGHT
    elif issubclass(obj_class, TrafficBarrier) or obj_class is TrafficBarrier:
        return MetaDriveType.TRAFFIC_BARRIER
    elif issubclass(obj_class, TrafficCone) or obj_class is TrafficCone:
        return MetaDriveType.TRAFFIC_CONE
    else:
        return MetaDriveType.OTHER


def _convert_type_to_string(nested):
    if isinstance(nested, type):
        return (nested.__module__, nested.__name__)
    if isinstance(nested, (list, tuple)):
        return [_convert_type_to_string(v) for v in nested]
    if isinstance(nested, dict):
        return {k: _convert_type_to_string(v) for k, v in nested.items()}
    return nested


def find_light_manager_name(manager_info):
    """
    Find the light_manager in real data manager
    """
    for manager_name in manager_info:
        if "LightManager" in manager_name:
            return manager_name
    return None


def find_traffic_manager_name(manager_info):
    """
    Find the traffic_manager in real data manager
    """
    for manager_name in manager_info:
        if "TrafficManager" in manager_name and manager_name != "PGTrafficManager":
            return manager_name
    return None


def find_data_manager_name(manager_info):
    """
    Find the data_manager
    """
    for manager_name in manager_info:
        if "DataManager" in manager_name and "NuPlan" not in manager_name:
            # Raw data in NuplanDataManager can not be dumped
            return manager_name
    return None


def convert_recorded_scenario_exported(record_episode, scenario_log_interval=0.1, to_dict=True):
    """
    This function utilizes the recorded data natively emerging from MetaDrive run.
    The output data structure follows MetaDrive data format, but some changes might happen compared to original data.
    For example, MetaDrive InterpolateLane will reformat the Lane data and making all waypoints equal distancing.
    We call this lane sampling rate, which is 0.2m in MetaDrive but might different in other dataset.
    """
    result = SD()

    result[SD.ID] = "{}-{}".format(record_episode["map_data"]["map_type"], record_episode["scenario_index"])

    result[SD.VERSION] = DATA_VERSION

    result["map_features"] = record_episode["map_data"]["map_features"]

    result[SD.DYNAMIC_MAP_STATES] = {}

    # TODO: Fix this
    if scenario_log_interval != 0.1:
        raise ValueError("We don't support varying the scenario log interval yet.")

    frames = [step_frame_list[-1] for step_frame_list in record_episode["frame"]]

    episode_len = len(frames)
    assert frames[-1].episode_step == episode_len - 1, "Length mismatch"
    result[SD.LENGTH] = episode_len

    result[SD.METADATA] = {}
    result[SD.METADATA][SD.METADRIVE_PROCESSED] = True
    result[SD.METADATA][SD.ID] = result[SD.ID]
    result[SD.METADATA]["dataset"] = "metadrive"
    result[SD.METADATA]["seed"] = record_episode["global_seed"]
    result[SD.METADATA]["scenario_id"] = record_episode["scenario_index"]
    result[SD.METADATA][SD.COORDINATE] = MetaDriveType.COORDINATE_METADRIVE
    result[SD.METADATA][SD.SDC_ID] = str(frames[0]._agent_to_object[DEFAULT_AGENT])
    result[SD.METADATA][SD.TIMESTEP] = \
        np.asarray([scenario_log_interval * i for i in range(episode_len)], dtype=np.float32)

    agent_to_object = {}
    object_to_agent = {}

    # Fill tracks
    all_objs = set()
    for frame in frames:
        all_objs.update(frame.step_info.keys())

    traffic_manager_name = find_traffic_manager_name(record_episode["manager_metadata"])
    light_manager_name = find_light_manager_name(record_episode["manager_metadata"])
    data_manager_name = find_data_manager_name(record_episode["manager_metadata"])

    tracks = {
        k: dict(
            type=MetaDriveType.UNSET,
            state=dict(
                position=np.zeros(shape=(episode_len, 3)),
                heading=np.zeros(shape=(episode_len, )),
                velocity=np.zeros(shape=(episode_len, 2)),
                valid=np.zeros(shape=(episode_len, )),

                # Add these items when the object has them.
                # throttle_brake=np.zeros(shape=(episode_len, 1)),
                # steering=np.zeros(shape=(episode_len, 1)),
                # size=np.zeros(shape=(episode_len, 3)),
            ),
            metadata=dict(track_length=episode_len, type=MetaDriveType.UNSET, object_id=k, original_id=k)
        )
        for k in list(all_objs)
    }

    all_lights = set()
    if light_manager_name is not None:
        for frame in frames:
            all_lights.update(frame.manager_info[light_manager_name][SD.ORIGINAL_ID_TO_OBJ_ID].keys())

    lights = {
        k: {
            "type": MetaDriveType.TRAFFIC_LIGHT,
            "state": {
                ScenarioDescription.TRAFFIC_LIGHT_STATUS: [MetaDriveType.LIGHT_UNKNOWN] * episode_len
            },
            ScenarioDescription.TRAFFIC_LIGHT_POSITION: np.zeros(shape=(3, ), dtype=np.float32),
            ScenarioDescription.TRAFFIC_LIGHT_LANE: None,
            "metadata": dict(
                track_length=episode_len, type=MetaDriveType.TRAFFIC_LIGHT, object_id=k, dataset="metadrive"
            )
        }
        for k in list(all_lights)
    }

    for frame_idx in range(result[SD.LENGTH]):

        # Record all agents' states (position, velocity, ...)
        for id, state in frames[frame_idx].step_info.items():
            # Fill type
            type = get_type_from_class(state["type"])
            if type == MetaDriveType.TRAFFIC_LIGHT:
                # pop id from tracks
                if id in tracks:
                    tracks.pop(id)
                assert light_manager_name in frames[frame_idx].manager_info, "Can not find light manager info"

                # convert to original id
                id = frames[frame_idx].manager_info[light_manager_name][SD.OBJ_ID_TO_ORIGINAL_ID][id]

                lights[id]["type"] = type
                lights[id][SD.METADATA]["type"] = lights[id]["type"]

                # Introducing the state item
                light_status = state[ScenarioDescription.TRAFFIC_LIGHT_STATUS]
                lights[id]["state"][ScenarioDescription.TRAFFIC_LIGHT_STATUS][frame_idx] = light_status

                # if light_status != MetaDriveType.LIGHT_UNKNOWN:
                if lights[id][ScenarioDescription.TRAFFIC_LIGHT_LANE] is None:
                    lights[id][ScenarioDescription.TRAFFIC_LIGHT_LANE] = str(id)
                    lights[id][ScenarioDescription.TRAFFIC_LIGHT_POSITION
                               ] = state[ScenarioDescription.TRAFFIC_LIGHT_POSITION]
                else:
                    assert lights[id][ScenarioDescription.TRAFFIC_LIGHT_LANE] == str(id)
                    assert lights[id][ScenarioDescription.TRAFFIC_LIGHT_POSITION
                                      ] == state[ScenarioDescription.TRAFFIC_LIGHT_POSITION]

            else:
                tracks[id]["type"] = type
                tracks[id][SD.METADATA]["type"] = tracks[id]["type"]

                # Introducing the state item
                tracks[id]["state"]["position"][frame_idx] = state["position"]
                tracks[id]["state"]["heading"][frame_idx] = state["heading_theta"]
                tracks[id]["state"]["velocity"][frame_idx] = state["velocity"]
                tracks[id]["state"]["valid"][frame_idx] = 1

                if "throttle_brake" in state:
                    if "throttle_brake" not in tracks[id]["state"]:
                        tracks[id]["state"]["throttle_brake"] = np.zeros(shape=(episode_len, 1))
                    tracks[id]["state"]["throttle_brake"][frame_idx] = state["throttle_brake"]

                if "steering" in state:
                    if "steering" not in tracks[id]["state"]:
                        tracks[id]["state"]["steering"] = np.zeros(shape=(episode_len, 1))
                    tracks[id]["state"]["steering"][frame_idx] = state["steering"]

                if "length" in state:
                    if "length" not in tracks[id]["state"]:
                        tracks[id]["state"]["length"] = np.zeros(shape=(episode_len, 1))
                    tracks[id]["state"]["length"][frame_idx] = state["length"]

                if "width" in state:
                    if "width" not in tracks[id]["state"]:
                        tracks[id]["state"]["width"] = np.zeros(shape=(episode_len, 1))
                    tracks[id]["state"]["width"][frame_idx] = state["width"]

                if "height" in state:
                    if "height" not in tracks[id]["state"]:
                        tracks[id]["state"]["height"] = np.zeros(shape=(episode_len, 1))
                    tracks[id]["state"]["height"][frame_idx] = state["height"]

                if id in frames[frame_idx]._object_to_agent:
                    tracks[id]["metadata"]["agent_name"] = frames[frame_idx]._object_to_agent[id]

                if traffic_manager_name is not None:
                    origin_id = frames[frame_idx].manager_info[traffic_manager_name][SD.OBJ_ID_TO_ORIGINAL_ID][id]
                    if tracks[id]["metadata"]["original_id"] == id:
                        tracks[id]["metadata"]["original_id"] = origin_id
                    else:
                        assert tracks[id]["metadata"]["original_id"] == origin_id

        # Record all policies' states (action, ...)
        for id, policy_info in frames[frame_idx].policy_info.items():
            # Maybe actions is also recorded. If so, add item to tracks:
            # TODO: In the case of discrete action, what should we do?
            for key, policy_state in policy_info.items():
                if policy_state is {}:
                    continue
                policy_state = np.asarray(policy_state)
                assert policy_state.dtype != object
                if key not in tracks[id]["state"]:
                    tracks[id]["state"][key] = np.zeros(
                        shape=(episode_len, policy_state.size), dtype=policy_state.dtype
                    )
                tracks[id]["state"][key][frame_idx] = policy_state

        # Record policy metadata
        for id, policy_spawn_info in frames[frame_idx].policy_spawn_info.items():
            tracks[id]["metadata"]["policy_spawn_info"] = copy.deepcopy(_convert_type_to_string(policy_spawn_info))

        # Record agent2object, object2agent mapping for metadata
        agent_to_object.update(frames[frame_idx]._agent_to_object)
        object_to_agent.update(frames[frame_idx]._object_to_agent)

        # Record spawn information
        for obj_name, spawn_info in frames[frame_idx].spawn_info.items():
            if obj_name not in tracks:
                continue
            spawn_info = copy.deepcopy(spawn_info)
            spawn_info = _convert_type_to_string(spawn_info)
            if "config" in spawn_info:
                spawn_info.pop("config")
            tracks[obj_name]["metadata"]["spawn_info"] = spawn_info

    result[SD.TRACKS] = tracks
    result[SD.DYNAMIC_MAP_STATES] = lights

    # # Traffic Light: Straight-through forward from original data
    # for k, manager_state in record_episode["manager_metadata"].items():
    #     if "DataManager" in k:
    #         if "raw_data" in manager_state:
    #             original_dynamic_map = copy.deepcopy(manager_state["raw_data"][SD.DYNAMIC_MAP_STATES])
    #             clipped_dynamic_map = {}
    #             for obj_id, obj_state in original_dynamic_map.items():
    #                 obj_state["state"] = {k: v[:episode_len] for k, v in obj_state["state"].items()}
    #                 clipped_dynamic_map[obj_id] = obj_state
    #             result[SD.METADATA]["history_metadata"] = manager_state["raw_data"][SD.METADATA]
    #             result[SD.DYNAMIC_MAP_STATES] = clipped_dynamic_map

    # Record agent2object, object2agent metadata
    result[SD.METADATA]["agent_to_object"] = {str(k): str(v) for k, v in agent_to_object.items()}
    result[SD.METADATA]["object_to_agent"] = {str(k): str(v) for k, v in object_to_agent.items()}

    if data_manager_name is not None:
        data_manager_raw_data = record_episode["manager_metadata"][data_manager_name].get("raw_data", None)
        if data_manager_raw_data:
            result[SD.METADATA]["history_metadata"] = data_manager_raw_data["metadata"]
    if to_dict:
        result = result.to_dict()
        SD.sanity_check(result, check_self_type=True)

    return result


def read_scenario_data(file_path):
    assert SD.is_scenario_file(file_path), "File: {} is not scenario file".format(file_path)
    with open(file_path, "rb") as f:
        # unpickler = CustomUnpickler(f)
        data = pickle.load(f)
    data = ScenarioDescription(data)
    return data


def read_dataset_summary(file_folder, check_file_existence=True):
    """
    We now support two methods to load pickle files.

    The first is the old method where we store pickle files in 0.pkl, 1.pkl, ...

    The second is the new method which use a summary file to record important metadata of each scenario.
    """
    summary_file = os.path.join(file_folder, SD.DATASET.SUMMARY_FILE)
    mapping_file = os.path.join(file_folder, SD.DATASET.MAPPING_FILE)
    if os.path.isfile(summary_file):
        with open(summary_file, "rb") as f:
            summary_dict = pickle.load(f)

    else:
        # Create a fake one
        files = []
        for file in os.listdir(file_folder):
            if SD.is_scenario_file(os.path.basename(file)):
                files.append(file)
        try:
            files = sorted(files, key=lambda file_name: int(file_name.replace(".pkl", "")))
        except ValueError:
            files = sorted(files, key=lambda file_name: file_name.replace(".pkl", ""))
        files = [p for p in files]
        summary_dict = {f: read_scenario_data(os.path.join(file_folder, f))["metadata"] for f in files}

    if os.path.exists(mapping_file):
        with open(mapping_file, "rb") as f:
            mapping = pickle.load(f)
    else:
        # Create a fake one
        mapping = {k: "" for k in summary_dict}

    if check_file_existence:
        for file in summary_dict:
            assert file in mapping, "FileName in mapping mismatch with summary"
            assert SD.is_scenario_file(file), "File:{} is not sd scenario file".format(file)
            file_path = os.path.join(file_folder, mapping[file], file)
            assert os.path.exists(file_path), "Can not find file: {}".format(file_path)

    return summary_dict, list(summary_dict.keys()), mapping


def get_number_of_scenarios(dataset_path):
    _, files, _ = read_dataset_summary(dataset_path)
    return len(files)


def assert_scenario_equal(scenarios1, scenarios2, only_compare_sdc=False, check_self_type=True):
    # ===== These two set of data should align =====
    assert set(scenarios1.keys()) == set(scenarios2.keys())
    for scenario_id in scenarios1.keys():
        SD.sanity_check(scenarios1[scenario_id], check_self_type=check_self_type)
        SD.sanity_check(scenarios2[scenario_id], check_self_type=check_self_type)
        old_scene = SD(scenarios1[scenario_id])
        new_scene = SD(scenarios2[scenario_id])
        SD.sanity_check(old_scene)
        SD.sanity_check(new_scene)
        assert old_scene[SD.LENGTH] >= new_scene[SD.LENGTH], (old_scene[SD.LENGTH], new_scene[SD.LENGTH])

        if only_compare_sdc:
            sdc1 = old_scene[SD.METADATA][SD.SDC_ID]
            sdc2 = new_scene[SD.METADATA][SD.SDC_ID]
            state_dict1 = old_scene[SD.TRACKS][sdc1]
            state_dict2 = new_scene[SD.TRACKS][sdc2]
            min_len = min(state_dict1[SD.STATE]["position"].shape[0], state_dict2[SD.STATE]["position"].shape[0])
            max_len = max(state_dict1[SD.STATE]["position"].shape[0], state_dict2[SD.STATE]["position"].shape[0])
            assert min_len / max_len > MIN_LENGTH_RATIO, "Replayed Scenario length ratio: {}".format(min_len / max_len)
            for k in state_dict1[SD.STATE].keys():
                if k in ["action", "throttle_brake", "steering"]:
                    continue
                elif k == "position":
                    np.testing.assert_almost_equal(
                        state_dict1[SD.STATE][k][:min_len][..., :2],
                        state_dict2[SD.STATE][k][:min_len][..., :2],
                        decimal=NP_ARRAY_DECIMAL
                    )
                elif k == "heading":
                    np.testing.assert_almost_equal(
                        wrap_to_pi(state_dict1[SD.STATE][k][:min_len] - state_dict2[SD.STATE][k][:min_len]),
                        np.zeros_like(state_dict2[SD.STATE][k][:min_len]),
                        decimal=NP_ARRAY_DECIMAL
                    )
                elif k == "velocity":
                    np.testing.assert_almost_equal(
                        state_dict1[SD.STATE][k][:min_len],
                        state_dict2[SD.STATE][k][:min_len],
                        decimal=VELOCITY_DECIMAL
                    )
            assert state_dict1[SD.TYPE] == state_dict2[SD.TYPE]

        else:
            # assert set(old_scene[SD.TRACKS].keys()).issuperset(set(new_scene[SD.TRACKS].keys()) - {new_scene[SD.METADATA][SD.SDC_ID]})
            assert len(old_scene[SD.TRACKS]) == len(new_scene[SD.TRACKS]), "obj num mismatch"
            for track_id, track in old_scene[SD.TRACKS].items():
                if track_id == new_scene[SD.METADATA][SD.SDC_ID]:
                    continue
                if track_id not in new_scene[SD.TRACKS]:
                    assert track_id == old_scene[SD.METADATA][SD.SDC_ID]
                    continue
                for state_k in new_scene[SD.TRACKS][track_id][SD.STATE]:
                    state_array_1 = new_scene[SD.TRACKS][track_id][SD.STATE][state_k]
                    state_array_2 = track[SD.STATE][state_k]
                    min_len = min(state_array_1.shape[0], state_array_2.shape[0])
                    max_len = max(state_array_1.shape[0], state_array_2.shape[0])
                    assert min_len / max_len > MIN_LENGTH_RATIO, "Replayed Scenario length ratio: {}".format(
                        min_len / max_len
                    )

                    if state_k == "velocity":
                        decimal = VELOCITY_DECIMAL
                    else:
                        decimal = NP_ARRAY_DECIMAL

                    if state_k == "heading":
                        # error < 5.7 degree is acceptable
                        broader_ratio = 1
                        ret = abs(wrap_to_pi(state_array_1[:min_len] - state_array_2[:min_len])) < 1e-1
                        ratio = np.sum(np.asarray(ret, dtype=np.int8)) / len(ret)
                        if ratio < broader_ratio:
                            raise ValueError("Match ration: {}, Target: {}".format(ratio, broader_ratio))

                        strict_ratio = 0.98
                        ret = abs(wrap_to_pi(state_array_1[:min_len] - state_array_2[:min_len])) < 1e-4
                        ratio = np.sum(np.asarray(ret, dtype=np.int8)) / len(ret)
                        if ratio < strict_ratio:
                            raise ValueError("Match ration: {}, Target: {}".format(ratio, strict_ratio))
                    else:
                        strict_ratio = 0.99
                        ret = abs(wrap_to_pi(state_array_1[:min_len] - state_array_2[:min_len])) < pow(10, -decimal)
                        ratio = np.sum(np.asarray(ret, dtype=np.int8)) / len(ret)
                        if ratio < strict_ratio:
                            raise ValueError("Match ration: {}, Target: {}".format(ratio, strict_ratio))

                assert new_scene[SD.TRACKS][track_id][SD.TYPE] == track[SD.TYPE]

            track_id = new_scene[SD.METADATA][SD.SDC_ID]
            for k in new_scene.get_sdc_track()["state"]:
                state_array_1 = new_scene.get_sdc_track()["state"][k]
                state_array_2 = old_scene.get_sdc_track()["state"][k]
                min_len = min(state_array_1.shape[0], state_array_2.shape[0])
                max_len = max(state_array_1.shape[0], state_array_2.shape[0])
                assert min_len / max_len > MIN_LENGTH_RATIO, "Replayed Scenario length ratio: {}".format(
                    min_len / max_len
                )

                if k == "velocity":
                    decimal = VELOCITY_DECIMAL
                elif k == "position":
                    state_array_1 = state_array_1[..., :2]
                    state_array_2 = state_array_2[..., :2]
                    decimal = NP_ARRAY_DECIMAL
                else:
                    decimal = NP_ARRAY_DECIMAL
                np.testing.assert_almost_equal(state_array_1[:min_len], state_array_2[:min_len], decimal=decimal)

            assert new_scene[SD.TRACKS][track_id][SD.TYPE] == track[SD.TYPE]

        assert set(old_scene[SD.MAP_FEATURES].keys()).issuperset(set(new_scene[SD.MAP_FEATURES].keys()))
        assert set(old_scene[SD.DYNAMIC_MAP_STATES].keys()) == set(new_scene[SD.DYNAMIC_MAP_STATES].keys())

        for map_id, map_feat in new_scene[SD.MAP_FEATURES].items():
            # It is possible that some line are not included in new scene but exist in old scene.
            # old_scene_polyline = map_feat["polyline"]
            # if coordinate_transform:
            #     old_scene_polyline = waymo_to_metadrive_vector(old_scene_polyline)
            np.testing.assert_almost_equal(
                new_scene[SD.MAP_FEATURES][map_id]["polyline"], map_feat["polyline"], decimal=NP_ARRAY_DECIMAL
            )
            assert new_scene[SD.MAP_FEATURES][map_id][SD.TYPE] == map_feat[SD.TYPE]

        for obj_id, obj_state in old_scene[SD.DYNAMIC_MAP_STATES].items():
            new_state_dict = new_scene[SD.DYNAMIC_MAP_STATES][obj_id][SD.STATE]
            old_state_dict = obj_state[SD.STATE]
            assert set(new_state_dict.keys()) == set(old_state_dict.keys())
            for k in new_state_dict.keys():
                min_len = min(new_state_dict[k].shape[0], old_state_dict[k].shape[0])
                max_len = max(new_state_dict[k].shape[0], old_state_dict[k].shape[0])
                assert min_len / max_len > MIN_LENGTH_RATIO, "Replayed Scenario length ratio: {}".format(
                    min_len / max_len
                )
                if k == ScenarioDescription.TRAFFIC_LIGHT_STATUS:
                    same_light = new_state_dict[k][:min_len] == old_state_dict[k][:min_len]
                    assert same_light.all()
                else:
                    np.testing.assert_almost_equal(
                        new_state_dict[k][:min_len], old_state_dict[k][:min_len], decimal=NP_ARRAY_DECIMAL
                    )

            assert new_scene[SD.DYNAMIC_MAP_STATES][obj_id][SD.TYPE] == obj_state[SD.TYPE]



"""

RenderPipeline

Copyright (c) 2014-2016 tobspr <tobias.springer1@gmail.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

"""

from panda3d.core import Vec3

from metadrive.render_pipeline.rpcore.image import Image
from metadrive.render_pipeline.rpcore.gui.draggable_window import DraggableWindow
from metadrive.render_pipeline.rpcore.gui.sprite import Sprite
from metadrive.render_pipeline.rpcore.gui.text import Text
from metadrive.render_pipeline.rpcore.gui.slider import Slider
from metadrive.render_pipeline.rpcore.gui.labeled_checkbox import LabeledCheckbox

from metadrive.render_pipeline.rpcore.util.display_shader_builder import DisplayShaderBuilder


class TexturePreview(DraggableWindow):
    """ Small window which provides a preview of a texture """
    def __init__(self, pipeline, parent):
        DraggableWindow.__init__(self, width=1600, height=900, parent=parent, title="Texture Viewer")
        self._pipeline = pipeline
        self._current_tex = None
        self._mip_slider = None
        self._mip_text = None
        self._slice_slider = None
        self._slice_text = None
        self._preview_image = None
        self._create_components()

    def present(self, tex):
        """ "Presents" a given texture and shows the window """
        self._current_tex = tex

        self.set_title(tex.get_name())

        # tex.write(tex.get_name() + ".png")

        # Remove old content
        self._content_node.node().remove_all_children()

        w, h = tex.get_x_size(), tex.get_y_size()
        if h > 1:
            scale_x = (self._width - 40.0) / w
            scale_y = (self._height - 110.0) / h
            scale_f = min(scale_x, scale_y)
            display_w = scale_f * w
            display_h = scale_f * h

        else:
            display_w = self._width - 40
            display_h = self._height - 110

        image = Sprite(
            image=tex,
            parent=self._content_node,
            x=20,
            y=90,
            w=display_w,
            h=display_h,
            any_filter=False,
            transparent=False
        )
        description = ""

        # Image size
        description += "{:d} x {:d} x {:d}".format(tex.get_x_size(), tex.get_y_size(), tex.get_z_size())

        # Image type
        description += ", {:s}, {:s}".format(
            Image.format_format(tex.get_format()).upper(),
            Image.format_component_type(tex.get_component_type()).upper()
        )

        Text(text=description, parent=self._content_node, x=17, y=70, size=16, color=Vec3(0.6, 0.6, 0.6))

        estimated_bytes = tex.estimate_texture_memory()
        size_desc = "Estimated memory: {:2.2f} MB".format(estimated_bytes / (1024.0**2))

        Text(
            text=size_desc,
            parent=self._content_node,
            x=self._width - 20.0,
            y=70,
            size=18,
            color=Vec3(0.34, 0.564, 0.192),
            align="right"
        )

        x_pos = len(size_desc) * 9 + 140

        # Slider for viewing different mipmaps
        if tex.uses_mipmaps():
            max_mips = tex.get_expected_num_mipmap_levels() - 1
            self._mip_slider = Slider(
                parent=self._content_node,
                size=140,
                min_value=0,
                max_value=max_mips,
                callback=self._set_mip,
                x=x_pos,
                y=65,
                value=0
            )
            x_pos += 140 + 5

            self._mip_text = Text(
                text="MIP: 5", parent=self._content_node, x=x_pos, y=72, size=18, color=Vec3(1, 0.4, 0.4), may_change=1
            )
            x_pos += 50 + 30

        # Slider for viewing different Z-layers
        if tex.get_z_size() > 1:
            self._slice_slider = Slider(
                parent=self._content_node,
                size=250,
                min_value=0,
                max_value=tex.get_z_size() - 1,
                callback=self._set_slice,
                x=x_pos,
                y=65,
                value=0
            )
            x_pos += 250 + 5

            self._slice_text = Text(
                text="Z: 5", parent=self._content_node, x=x_pos, y=72, size=18, color=Vec3(0.4, 1, 0.4), may_change=1
            )

            x_pos += 50 + 30

        # Slider to adjust brightness
        self._bright_slider = Slider(
            parent=self._content_node,
            size=140,
            min_value=-14,
            max_value=14,
            callback=self._set_brightness,
            x=x_pos,
            y=65,
            value=0
        )
        x_pos += 140 + 5
        self._bright_text = Text(
            text="Bright: 1", parent=self._content_node, x=x_pos, y=72, size=18, color=Vec3(0.4, 0.4, 1), may_change=1
        )
        x_pos += 100 + 30

        # Slider to enable reinhard tonemapping
        self._tonemap_box = LabeledCheckbox(
            parent=self._content_node,
            x=x_pos,
            y=60,
            text="Tonemap",
            text_color=Vec3(1, 0.4, 0.4),
            chb_checked=False,
            chb_callback=self._set_enable_tonemap,
            text_size=18,
            expand_width=90
        )
        x_pos += 90 + 30

        image.set_shader_inputs(slice=0, mipmap=0, brightness=1, tonemap=False)

        preview_shader = DisplayShaderBuilder.build(tex, display_w, display_h)
        image.set_shader(preview_shader)

        self._preview_image = image
        self.show()

    def _set_slice(self):
        idx = int(self._slice_slider.value)
        self._preview_image.set_shader_input("slice", idx)
        self._slice_text.set_text("Z: " + str(idx))

    def _set_mip(self):
        idx = int(self._mip_slider.value)
        self._preview_image.set_shader_input("mipmap", idx)
        self._mip_text.set_text("MIP " + str(idx))

    def _set_brightness(self):
        val = self._bright_slider.value
        scale = 2**val
        self._bright_text.set_text("Bright: " + str(round(scale, 3)))
        self._preview_image.set_shader_input("brightness", scale)

    def _set_enable_tonemap(self, enable_tonemap):
        self._preview_image.set_shader_input("tonemap", enable_tonemap)

    def _create_components(self):
        """ Internal method to init the components """
        DraggableWindow._create_components(self)
        self._content_node = self._node.attach_new_node("content")



import matplotlib.pyplot as plt
from metadrive.scenario.utils import read_scenario_data
from matplotlib.pyplot import figure

from metadrive.type import MetaDriveType
from metadrive.utils.math import mph_to_kmh
from metadrive.utils.waymo.waymo_type import WaymoLaneType, WaymoAgentType
from metadrive.utils.waymo.waymo_type import WaymoRoadLineType, WaymoRoadEdgeType

try:
    import tensorflow as tf
except ImportError:
    pass
try:
    from metadrive.utils.waymo.protos import scenario_pb2
except ImportError:
    pass
import pickle
import numpy as np
from metadrive.scenario.scenario_description import ScenarioDescription

import logging

logger = logging.getLogger(__name__)

logger.warning(
    "This converters will de deprecated. "
    "Use tools in ScenarioNet instead: https://github.com/metadriverse/ScenarioNet."
)


def extract_poly(message):
    x = [i.x for i in message]
    y = [i.y for i in message]
    z = [i.z for i in message]
    coord = np.stack((x, y, z), axis=1).astype("float32")
    return coord


def extract_boundaries(fb):
    b = []
    # b = np.zeros([len(fb), 4], dtype="int64")
    for k in range(len(fb)):
        c = dict()
        c["lane_start_index"] = fb[k].lane_start_index
        c["lane_end_index"] = fb[k].lane_end_index
        c["boundary_type"] = WaymoRoadLineType.from_waymo(fb[k].boundary_type)
        c["boundary_feature_id"] = fb[k].boundary_feature_id
        for key in c:
            c[key] = str(c[key])
        b.append(c)

    return b


def extract_neighbors(fb):
    nbs = []
    for k in range(len(fb)):
        nb = dict()
        nb["feature_id"] = fb[k].feature_id
        nb["self_start_index"] = fb[k].self_start_index
        nb["self_end_index"] = fb[k].self_end_index
        nb["neighbor_start_index"] = fb[k].neighbor_start_index
        nb["neighbor_end_index"] = fb[k].neighbor_end_index
        for key in nb:
            nb[key] = str(nb[key])
        nb["boundaries"] = extract_boundaries(fb[k].boundaries)
        nbs.append(nb)
    return nbs


def extract_center(f):
    center = dict()
    f = f.lane
    center["speed_limit_mph"] = f.speed_limit_mph

    center["speed_limit_kmh"] = mph_to_kmh(f.speed_limit_mph)

    center["type"] = WaymoLaneType.from_waymo(f.type)

    center["polyline"] = extract_poly(f.polyline)

    center["interpolating"] = f.interpolating

    center["entry_lanes"] = [x for x in f.entry_lanes]

    center["exit_lanes"] = [x for x in f.exit_lanes]

    center["left_boundaries"] = extract_boundaries(f.left_boundaries)

    center["right_boundaries"] = extract_boundaries(f.right_boundaries)

    center["left_neighbor"] = extract_neighbors(f.left_neighbors)

    center["right_neighbor"] = extract_neighbors(f.right_neighbors)

    return center


def extract_line(f):
    line = dict()
    f = f.road_line
    line["type"] = WaymoRoadLineType.from_waymo(f.type)
    line["polyline"] = extract_poly(f.polyline)
    return line


def extract_edge(f):
    edge = dict()
    f_ = f.road_edge

    # TODO: Need to transform this to MetaDrive version
    edge["type"] = WaymoRoadEdgeType.from_waymo(f_.type)

    edge["polyline"] = extract_poly(f_.polyline)

    return edge


def extract_stop(f):
    stop = dict()
    f = f.stop_sign
    stop["type"] = MetaDriveType.STOP_SIGN
    stop["lane"] = [x for x in f.lane]
    stop["position"] = np.array([f.position.x, f.position.y, f.position.z], dtype="float32")
    return stop


def extract_crosswalk(f):
    cross_walk = dict()
    f = f.crosswalk
    cross_walk["type"] = MetaDriveType.CROSSWALK
    cross_walk["polygon"] = extract_poly(f.polygon)
    return cross_walk


def extract_bump(f):
    speed_bump_data = dict()
    f = f.speed_bump
    speed_bump_data["type"] = MetaDriveType.SPEED_BUMP
    speed_bump_data["polygon"] = extract_poly(f.polygon)
    return speed_bump_data


def extract_driveway(f):
    driveway_data = dict()
    f = f.driveway
    driveway_data["type"] = MetaDriveType.DRIVEWAY
    driveway_data["polygon"] = extract_poly(f.polygon)
    return driveway_data


def extract_tracks(tracks, sdc_idx, track_length):
    ret = dict()

    def _object_state_template(object_id):
        return dict(
            type=None,
            state=dict(

                # Never add extra dim if the value is scalar.
                position=np.zeros([track_length, 3], dtype=np.float32),
                length=np.zeros([track_length], dtype=np.float32),
                width=np.zeros([track_length], dtype=np.float32),
                height=np.zeros([track_length], dtype=np.float32),
                heading=np.zeros([track_length], dtype=np.float32),
                velocity=np.zeros([track_length, 2], dtype=np.float32),
                valid=np.zeros([track_length], dtype=bool),
            ),
            metadata=dict(track_length=track_length, type=None, object_id=object_id, dataset="waymo")
        )

    for obj in tracks:
        object_id = str(obj.id)

        obj_state = _object_state_template(object_id)

        waymo_string = WaymoAgentType.from_waymo(obj.object_type)  # Load waymo type string
        metadrive_type = MetaDriveType.from_waymo(waymo_string)  # Transform it to Waymo type string
        obj_state["type"] = metadrive_type

        for step_count, state in enumerate(obj.states):

            if step_count >= track_length:
                break

            obj_state["state"]["position"][step_count][0] = state.center_x
            obj_state["state"]["position"][step_count][1] = state.center_y
            obj_state["state"]["position"][step_count][2] = state.center_z

            # l = [state.length for state in obj.states]
            # w = [state.width for state in obj.states]
            # h = [state.height for state in obj.states]
            # obj_state["state"]["size"] = np.stack([l, w, h], 1).astype("float32")
            obj_state["state"]["length"][step_count] = state.length
            obj_state["state"]["width"][step_count] = state.width
            obj_state["state"]["height"][step_count] = state.height

            # heading = [state.heading for state in obj.states]
            obj_state["state"]["heading"][step_count] = state.heading

            obj_state["state"]["velocity"][step_count][0] = state.velocity_x
            obj_state["state"]["velocity"][step_count][1] = state.velocity_y

            obj_state["state"]["valid"][step_count] = state.valid

        obj_state["metadata"]["type"] = metadrive_type

        ret[object_id] = obj_state

    return ret, str(tracks[sdc_idx].id)


def extract_map_features(map_features):
    ret = {}

    for lane_state in map_features:
        lane_id = str(lane_state.id)

        if lane_state.HasField("lane"):
            ret[lane_id] = extract_center(lane_state)

        if lane_state.HasField("road_line"):
            ret[lane_id] = extract_line(lane_state)

        if lane_state.HasField("road_edge"):
            ret[lane_id] = extract_edge(lane_state)

        if lane_state.HasField("stop_sign"):
            ret[lane_id] = extract_stop(lane_state)

        if lane_state.HasField("crosswalk"):
            ret[lane_id] = extract_crosswalk(lane_state)

        if lane_state.HasField("speed_bump"):
            ret[lane_id] = extract_bump(lane_state)

        # Supported only in Waymo dataset 1.2.0
        if lane_state.HasField("driveway"):
            ret[lane_id] = extract_driveway(lane_state)

    return ret


def extract_dynamic_map_states(dynamic_map_states, track_length):
    processed_dynamics_map_states = {}

    def _traffic_light_state_template(object_id):
        return dict(
            type=MetaDriveType.TRAFFIC_LIGHT,
            state=dict(object_state=[None] * track_length),
            lane=None,
            stop_point=np.zeros([
                3,
            ], dtype=np.float32),
            metadata=dict(
                track_length=track_length, type=MetaDriveType.TRAFFIC_LIGHT, object_id=object_id, dataset="waymo"
            )
        )

    for step_count, step_states in enumerate(dynamic_map_states):
        # Each step_states is the state of all objects in one time step
        lane_states = step_states.lane_states

        if step_count >= track_length:
            break

        for object_state in lane_states:
            lane = object_state.lane
            object_id = str(lane)  # Always use string to specify object id

            # We will use lane index to serve as the traffic light index.
            if object_id not in processed_dynamics_map_states:
                processed_dynamics_map_states[object_id] = _traffic_light_state_template(object_id=object_id)

            if processed_dynamics_map_states[object_id]["lane"] is not None:
                assert lane == processed_dynamics_map_states[object_id]["lane"]
            else:
                processed_dynamics_map_states[object_id]["lane"] = lane

            object_state_string = object_state.State.Name(object_state.state)
            processed_dynamics_map_states[object_id]["state"]["object_state"][step_count] = object_state_string

            processed_dynamics_map_states[object_id]["stop_point"][0] = object_state.stop_point.x
            processed_dynamics_map_states[object_id]["stop_point"][1] = object_state.stop_point.y
            processed_dynamics_map_states[object_id]["stop_point"][2] = object_state.stop_point.z

    for obj in processed_dynamics_map_states.values():
        assert len(obj["state"]["object_state"]) == obj["metadata"]["track_length"]

    return processed_dynamics_map_states


class CustomUnpickler(pickle.Unpickler):
    def __init__(self, load_old_scenario, *args, **kwargs):
        raise DeprecationWarning("Now we don't pickle any customized data type, so this class is deprecated now")
        super(CustomUnpickler, self).__init__(*args, **kwargs)
        self.load_old_scenario = load_old_scenario

    def find_class(self, module, name):
        if self.load_old_scenario:
            raise ValueError("Old scenario is completely deprecated. Can't load it any more.")
            if name == "AgentType":
                return AgentTypeClass
            elif name == "RoadLineType":
                return RoadLineTypeClass
            elif name == "RoadEdgeType":
                return RoadEdgeTypeClass
            return super().find_class(module, name)
        else:
            return super().find_class(module, name)


def read_waymo_data(file_path):
    return read_scenario_data(file_path)


def draw_waymo_map(data):
    """
    TODO: Need this function in future.
    """
    figure(figsize=(8, 6), dpi=500)
    for key, value in data[ScenarioDescription.MAP_FEATURES].items():
        if value.get("type", None) == "center_lane":
            plt.scatter([x[0] for x in value["polyline"]], [y[1] for y in value["polyline"]], s=0.5)
        elif value.get("type", None) == "road_edge":
            plt.scatter([x[0] for x in value["polyline"]], [y[1] for y in value["polyline"]], s=0.5, c=(0, 0, 0))
        # elif value.get("type", None) == "road_line":
        #     plt.scatter([x[0] for x in value["polyline"]], [y[1] for y in value["polyline"]], s=0.5, c=(0.8,0.8,0.8))
    plt.show()


# return the nearest point"s index of the line
def nearest_point(point, line):
    dist = np.square(line - point)
    dist = np.sqrt(dist[:, 0] + dist[:, 1])
    return np.argmin(dist)


def extract_width(map, polyline, boundary):
    l_width = np.zeros(polyline.shape[0], dtype="float32")
    for b in boundary:
        boundary_int = {k: int(v) if k != "boundary_type" else v for k, v in b.items()}  # All values are int

        b_feat_id = str(boundary_int["boundary_feature_id"])
        lb = map[b_feat_id]
        b_polyline = lb["polyline"][:, :2]

        start_p = polyline[boundary_int["lane_start_index"]]
        start_index = nearest_point(start_p, b_polyline)
        seg_len = boundary_int["lane_end_index"] - boundary_int["lane_start_index"]
        end_index = min(start_index + seg_len, lb["polyline"].shape[0] - 1)
        length = min(end_index - start_index, seg_len) + 1
        self_range = range(boundary_int["lane_start_index"], boundary_int["lane_start_index"] + length)
        bound_range = range(start_index, start_index + length)
        centerLane = polyline[self_range]
        bound = b_polyline[bound_range]
        dist = np.square(centerLane - bound)
        dist = np.sqrt(dist[:, 0] + dist[:, 1])
        l_width[self_range] = dist
    return l_width


def compute_width(map):
    for map_feat_id, lane in map.items():

        if not "LANE" in lane["type"]:
            continue

        width = np.zeros((lane["polyline"].shape[0], 2), dtype="float32")

        width[:, 0] = extract_width(map, lane["polyline"][:, :2], lane["left_boundaries"])
        width[:, 1] = extract_width(map, lane["polyline"][:, :2], lane["right_boundaries"])

        width[width[:, 0] == 0, 0] = width[width[:, 0] == 0, 1]
        width[width[:, 1] == 0, 1] = width[width[:, 1] == 0, 0]

        lane["width"] = width
    return


# parse raw data from input path to output path

# def convert_polyline_to_metadrive(waymo_polyline, coordinate_transform=True):
#     """
#     Waymo lane is in a different coordinate system, using them after converting
#     """
#     convert_polyline_to_metadrive(waymo_polyline, coordinate_transform)



import pathlib
import pickle

from metadrive.component.map.argoverse_map import ArgoverseMap
from metadrive.envs.metadrive_env import MetaDriveEnv
from metadrive.manager.pg_map_manager import PGMapManager
from metadrive.utils import is_win

argoverse_city = "PIT"
argoverse_map_center = [2599.5505965123866, 1200.0214763629717]
argoverse_map_radius = 150
argoverse_spawn_lane_index = ('7903', '9713', 0)
argoverse_destination = "968"
argoverse_log_id = "c6911883-1843-3727-8eaa-41dc8cda8993"


class ArgoversePGMapManager(PGMapManager):
    def before_reset(self):
        # do not unload map
        pass

    def reset(self):
        if self.current_map is None:
            self.engine.global_config["map_config"].update(
                {
                    "city": argoverse_city,
                    "center": ArgoverseMap.metadrive_vector(argoverse_map_center),
                    "radius": argoverse_map_radius
                }
            )
            map = ArgoverseMap(self.engine.global_config["map_config"])
            self.engine.map_manager.load_map(map)


class ArgoverseEnv(MetaDriveEnv):
    raise DeprecationWarning("Due to poor data, we are not maintaining this environment now!")

    def _post_process_config(self, config):
        config = super(ArgoverseEnv, self)._post_process_config(config)
        config["vehicle_config"]["spawn_lane_index"] = argoverse_spawn_lane_index
        config["vehicle_config"]["destination"] = argoverse_destination

        log_id = argoverse_log_id
        root_path = pathlib.PurePosixPath(__file__).parent.parent if not is_win() else pathlib.Path(__file__).resolve(
        ).parent.parent
        data_path = root_path.joinpath("assets").joinpath("real_data").joinpath("{}.pkl".format(log_id))
        with open(data_path, 'rb') as f:
            locate_info, _ = pickle.load(f)
        config.update({"real_data_config": {"locate_info": locate_info}})
        config["traffic_density"] = 0.0  # Remove rule-based traffic flow
        return config

    def setup_engine(self):
        super(ArgoverseEnv, self).setup_engine()
        from metadrive.manager.real_data_manager import RealDataManager
        self.engine.register_manager("real_data_manager", RealDataManager())
        self.engine.update_manager("map_manager", ArgoversePGMapManager())



import distutils.spawn
import distutils.version
import os
import subprocess
import time

import numpy as np
from gymnasium import logger, error
from panda3d.core import PNMImage

from metadrive.component.algorithm.BIG import BigGenerateMethod
from metadrive.component.map.base_map import BaseMap
from metadrive.envs.metadrive_env import MetaDriveEnv


class ImageEncoder(object):
    def __init__(self, output_path, frame_shape, frames_per_sec):
        self.proc = None
        self.output_path = output_path
        # Frame shape should be lines-first, so w and h are swapped
        h, w, pixfmt = frame_shape
        if pixfmt != 3 and pixfmt != 4:
            raise error.InvalidFrame(
                "Your frame has shape {}, but we require (w,h,3) or (w,h,4), "
                "i.e., RGB values for a w-by-h image, with an optional alpha "
                "channel.".format(frame_shape)
            )
        self.wh = (w, h)
        self.includes_alpha = (pixfmt == 4)
        self.frame_shape = frame_shape
        self.frames_per_sec = frames_per_sec

        if distutils.spawn.find_executable('avconv') is not None:
            self.backend = 'avconv'
        elif distutils.spawn.find_executable('ffmpeg') is not None:
            self.backend = 'ffmpeg'
        else:
            raise error.DependencyNotInstalled(
                """Found neither the ffmpeg nor avconv executables. On OS X, 
                you can install ffmpeg via `brew install ffmpeg`. On most 
                Ubuntu variants, `sudo apt-get install ffmpeg` should do it. 
                On Ubuntu 14.04, however, you'll need to install avconv with 
                `sudo apt-get install libav-tools`."""
            )

        self.start()

    @property
    def version_info(self):
        return {
            'backend': self.backend,
            'version': str(subprocess.check_output([self.backend, '-version'], stderr=subprocess.STDOUT)),
            'cmdline': self.cmdline
        }

    def start(self):
        self.cmdline = (
            self.backend,
            '-nostats',
            '-loglevel',
            'error',  # suppress warnings
            '-y',
            '-r',
            '%d' % self.frames_per_sec,
            # '-b', '2M',

            # input
            '-f',
            'rawvideo',
            '-s:v',
            '{}x{}'.format(*self.wh),
            '-pix_fmt',
            ('rgb32' if self.includes_alpha else 'rgb24'),
            '-i',
            '-',
            # this used to be /dev/stdin, which is not Windows-friendly

            # output
            '-vf',
            'scale=trunc(iw/2)*2:trunc(ih/2)*2',
            '-vcodec',
            'libx264',
            '-pix_fmt',
            'yuv420p',
            '-crf',
            '18',
            # '-vtag',
            # 'hvc1',
            self.output_path
        )

        logger.debug('Starting ffmpeg with "%s"', ' '.join(self.cmdline))
        if hasattr(os, 'setsid'):  # setsid not present on Windows
            self.proc = subprocess.Popen(self.cmdline, stdin=subprocess.PIPE, preexec_fn=os.setsid)
        else:
            self.proc = subprocess.Popen(self.cmdline, stdin=subprocess.PIPE)

    def capture_frame(self, frame):
        if not isinstance(frame, (np.ndarray, np.generic)):
            raise error.InvalidFrame(
                'Wrong type {} for {} (must be np.ndarray or np.generic)'.format(type(frame), frame)
            )
        if frame.shape != self.frame_shape:
            raise error.InvalidFrame(
                "Your frame has shape {}, but the VideoRecorder is "
                "configured for shape {}.".format(frame.shape, self.frame_shape)
            )
        if frame.dtype != np.uint8:
            raise error.InvalidFrame(
                "Your frame has data type {}, but we require uint8 (i.e. RGB "
                "values from 0-255).".format(frame.dtype)
            )

        if distutils.version.LooseVersion(np.__version__) >= distutils.version.LooseVersion('1.9.0'):
            self.proc.stdin.write(frame.tobytes())
        else:
            self.proc.stdin.write(frame.tostring())

    def close(self):
        self.proc.stdin.close()
        ret = self.proc.wait()
        if ret != 0:
            logger.error("VideoRecorder encoder exited with status {}".format(ret))


def gen_video(frames, file="tmp"):
    path = "{}.mp4".format(file)
    encoder = ImageEncoder(path, frames[0].shape, 50)
    for f in frames:
        encoder.capture_frame(f)
    encoder.close()
    del encoder
    print("Video is saved at: ", path)
    # video = io.open(path, 'r+b').read()
    # encoded = base64.b64encode(video)
    # ipythondisplay.display(HTML(data='''<video alt="test" autoplay
    #           loop controls style="height: 400px;">
    #           <source src="data:video/mp4;base64,{0}" type="video/mp4" />
    #         </video>'''.format(encoded.decode('ascii'))))


if __name__ == '__main__':
    env = MetaDriveEnv(
        dict(
            use_render=False,
            map_config={
                BaseMap.GENERATE_TYPE: BigGenerateMethod.BLOCK_NUM,
                BaseMap.GENERATE_CONFIG: 7
            },
            traffic_density=0.5,
            image_observation=True,
        )
    )
    start = time.time()
    env.reset()
    frames = []
    for num_frames in range(30):
        o, r, tm, tc, info = env.step([0, 1])
        img = PNMImage()
        env.engine.win.getScreenshot(img)
        frame = np.zeros([1200, 900, 4], dtype=np.uint8)
        for i in range(1200):
            for j in range(900):
                frame[i, j] = img.get_pixel(i, j)
        frame = frame.swapaxes(0, 1)[..., :3]
        frames.append(frame)
        # print(f"Finish {num_frames + 1} frames")
    env.close()
    gen_video(frames)



from metadrive.component.traffic_participants.pedestrian import Pedestrian
from metadrive.envs.metadrive_env import MetaDriveEnv


def test_pedestrian(render=False):
    env = MetaDriveEnv(
        {
            "num_scenarios": 1,
            "traffic_density": 0.,
            "traffic_mode": "hybrid",
            "start_seed": 22,
            "debug": False,
            "cull_scene": False,
            "manual_control": False,
            "use_render": render,
            "decision_repeat": 5,
            "need_inverse_traffic": False,
            "rgb_clip": True,
            "map": "X",
            # "agent_policy": IDMPolicy,
            "random_traffic": False,
            "random_lane_width": True,
            # "random_agent_model": True,
            "driving_reward": 1.0,
            "force_destroy": False,
            # "camera_dist": -1,
            # "camera_pitch": 30,
            # "camera_height": 1,
            # "camera_smooth": False,
            # "camera_height": -1,
            "window_size": (2400, 1600),
            "vehicle_config": {
                "enable_reverse": False,
            },
        }
    )
    env.reset()
    try:
        obj_1 = env.engine.spawn_object(Pedestrian, position=[30, 0], heading_theta=0, random_seed=1)
        obj_2 = env.engine.spawn_object(Pedestrian, position=[30, 6], heading_theta=0, random_seed=1)
        obj_1.set_velocity([1, 0], 1, in_local_frame=True)
        obj_2.set_velocity([1, 0], 0, in_local_frame=True)
        env.vehicle.set_velocity([5, 0], in_local_frame=False)
        for s in range(1, 1000):
            o, r, tm, tc, info = env.step([0, 0])
            # obj_1.set_velocity([1, 0], 2, in_local_frame=True)
            # obj_2.set_velocity([1, 0], 0.8, in_local_frame=True)
            if s == 300:
                obj_1.set_velocity([1, 0], 0, in_local_frame=True)
                # obj_2.set_velocity([1, 0], 0, in_local_frame=True)
            elif s == 500:
                obj_1.set_velocity([1, 0], 2, in_local_frame=True)

            # else:
            #     obj_1.set_velocity([1, 0], 1, in_local_frame=True)

            # if 100 < s < 300:
            #     obj_2.set_velocity([1, 0], 1, in_local_frame=True)
            # elif 500 > s > 300:
            #     obj_2.set_velocity([1, 0], 0, in_local_frame=True)
            # elif s >= 500:
            #     obj_2.set_velocity([1, 0], 2, in_local_frame=True)
        assert abs(obj_1.position[0] - 160) < 1, "Pedestrian movement error!"
    finally:
        env.close()


if __name__ == "__main__":
    test_pedestrian(True)



from metadrive.envs.marl_envs import MultiAgentTinyInter


class TestEnv(MultiAgentTinyInter):
    def __init__(self):
        super(TestEnv, self).__init__(
            config={

                # "num_agents": 8,
                # "map_config": {
                #     "exit_length": 30,
                #     "lane_num": 1,
                #     "lane_width": 4
                # },

                # === Debug ===
                "vehicle_config": {
                    "show_line_to_dest": True
                },
                "manual_control": True,
                # "num_agents": 4,
                "use_render": True,
            }
        )


if __name__ == "__main__":
    env = TestEnv()
    o, _ = env.reset()
    # print("vehicle num", len(env.engine.traffic_manager.vehicles))
    for i in range(1, 100000):
        o, r, tm, tc, info = env.step(env.action_space.sample())
        if True in tm.values() or True in tc.values():
            print("Somebody Done. ", info)
            # env.reset()
    env.close()



"""

RenderPipeline

Copyright (c) 2014-2016 tobspr <tobias.springer1@gmail.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

"""

import direct.gui.DirectGuiGlobals as DGG
from direct.gui.DirectSlider import DirectSlider

from metadrive.render_pipeline.rpcore.rpobject import RPObject


class Slider(RPObject):
    """ This is a simple wrapper around DirectSlider, providing a simpler
    interface """
    def __init__(
        self,
        x=0,
        y=0,
        parent=None,
        size=100,
        min_value=0,
        max_value=100,
        value=50,
        page_size=1,
        callback=None,
        extra_args=None
    ):
        """ Inits the slider """
        RPObject.__init__(self)
        if extra_args is None:
            extra_args = []

        # Scale has to be 2.0, otherwise there will be an error.
        self._node = DirectSlider(
            pos=(size * 0.5 + x, 1, -y),
            parent=parent,
            range=(min_value, max_value),
            value=value,
            pageSize=page_size,
            scale=2.0,
            command=callback,
            extraArgs=extra_args,
            frameColor=(0.0, 0.0, 0.0, 1),
            frameSize=(-size * 0.25, size * 0.25, -5, 5),
            relief=DGG.FLAT,
            thumb_frameColor=(0.35, 0.53, 0.2, 1.0),
            thumb_relief=DGG.FLAT,
            thumb_frameSize=(-2.5, 2.5, -5.0, 5.0),
        )

    @property
    def value(self):
        """ Returns the currently assigned value """
        return self._node["value"]

    @value.setter
    def value(self, value):
        """ Sets the value of the slider """
        self._node["value"] = value

    @property
    def node(self):
        """ Returns a handle to the internally used node """
        return self._node



# Copyright (c) 2020 kitsune.ONE team.

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from metadrive.third_party.kitsunetsuki.base.material import get_root_node, get_from_node

SHADING_MODEL_DEFAULT = 0
SHADING_MODEL_EMISSIVE = 1
SHADING_MODEL_CLEARCOAT = 2
SHADING_MODEL_TRANSPARENT_GLASS = 3
SHADING_MODEL_SKIN = 4
SHADING_MODEL_FOLIAGE = 5
SHADING_MODEL_TRANSPARENT_EMISSIVE = 6


class MaterialMixin(object):
    def make_material(self, material):
        gltf_material = {
            'name': material.name,
            'alphaMode': 'OPAQUE',
            'alphaCutoff': material.alpha_threshold,
            'doubleSided': not material.use_backface_culling,
            'pbrMetallicRoughness': {
                'extras': {},
            },
            'extras': {},
        }

        gltf_material['alphaMode'] = {
            'OPAQUE': 'OPAQUE',
            'BLEND': 'BLEND',
            'CLIP': 'MASK'
        }.get(material.blend_method, 'OPAQUE')

        shader = None
        if material.node_tree is not None:
            output = get_root_node(material.node_tree, 'OUTPUT_MATERIAL')
            shader = None
            if output:
                shader = get_from_node(
                    material.node_tree,
                    'BSDF_PRINCIPLED',
                    to_node=output,
                    from_socket_name='BSDF',
                    to_socket_name='Surface'
                )

        if not shader:
            return gltf_material

        if self._render_type == 'rp':  # RenderPipeline
            # emission = tuple(shader.inputs['Emission'].default_value)[:3]
            emission = (0, 0, 0)
            # alpha = shader.inputs['Alpha'].default_value
            alpha = 1
            clearcoat = shader.inputs['Clearcoat'].default_value
            normal_strength = self.get_normal_strength(material, shader)

            if sum(emission) > 0:  # emission
                gltf_material['pbrMetallicRoughness'].update(
                    {
                        'baseColorFactor': emission + (1, ),
                        'metallicFactor': 0,
                        'roughnessFactor': 1,
                    }
                )
                gltf_material['pbrMetallicRoughness']['extras']['ior'] = 1.51

                if alpha < 1:
                    emit = [SHADING_MODEL_TRANSPARENT_EMISSIVE, normal_strength, alpha]
                else:
                    emit = [SHADING_MODEL_EMISSIVE, normal_strength, 0]
                gltf_material['emissiveFactor'] = tuple(emit)

            else:  # not emission
                gltf_material['pbrMetallicRoughness'].update(
                    {
                        'baseColorFactor': (1, 1, 1, 1),
                        'metallicFactor': self.get_metallic(material, shader),
                        'roughnessFactor': self.get_roughness(material, shader),
                    }
                )
                gltf_material['pbrMetallicRoughness']['extras']['ior'] = shader.inputs['IOR'].default_value

                if alpha < 1:
                    emit = [SHADING_MODEL_TRANSPARENT_GLASS, normal_strength, alpha]
                elif clearcoat:
                    emit = [SHADING_MODEL_CLEARCOAT, normal_strength, 0]
                else:
                    emit = [SHADING_MODEL_DEFAULT, normal_strength, 0]
                gltf_material['emissiveFactor'] = tuple(emit)

        else:  # not RenderPipeline
            alpha = 1

            gltf_material['pbrMetallicRoughness'].update(
                {
                    'baseColorFactor': (1, 1, 1, alpha),
                    'metallicFactor': self.get_metallic(material, shader),
                    'roughnessFactor': self.get_roughness(material, shader),
                }
            )

            gltf_material['emissiveFactor'] = tuple(self.get_emission(material, shader))[:3]
            # if alpha < 1:
            #     gltf_material['alphaMode'] = 'BLEND'
            #     gltf_material['alphaCutoff'] = alpha
            # else:
            #     gltf_material['alphaMode'] = 'OPAQUE'
            #     gltf_material['alphaCutoff'] = 0

        if material.node_tree is not None:
            for node in material.node_tree.nodes:
                if node.type == 'ATTRIBUTE':
                    gltf_material['extras'][node.attribute_type.lower()] = node.attribute_name

        return gltf_material



# Copyright (c) 2020 kitsune.ONE team.

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


def get_armature(obj):
    parent = obj.parent
    while parent:
        if parent.type == 'ARMATURE':
            return parent
        parent = parent.parent


def is_left_bone(bone):
    return (
        bone.name.endswith('_L') or bone.name.endswith('.L') or bone.name.lower().startswith('left')
        or '.L.' in bone.name or '_L.' in bone.name or ':Left' in bone.name
    )


def is_bone_matches(bone, names):
    if bone is None:
        return False

    if bone.name.lower().endswith('_end'):
        return False

    for name in names:
        if name in bone.name.lower():
            return True

    return False



from metadrive.component.sensors.mini_map import MiniMap
from metadrive.component.sensors.rgb_camera import RGBCamera
from metadrive.component.sensors.depth_camera import DepthCamera
from metadrive.component.sensors.vehicle_panel import VehiclePanel
from metadrive.envs.safe_metadrive_env import SafeMetaDriveEnv

if __name__ == "__main__":

    def get_image(env):
        depth_cam = env.vehicle.get_camera(env.vehicle.config["image_source"])
        rgb_cam = env.vehicle.get_camera("rgb_camera")
        for h in range(-180, 180, 20):
            env.engine.graphicsEngine.renderFrame()
            depth_cam.get_cam().setH(h)
            rgb_cam.get_cam().setH(h)
            depth_cam.save_image(env.vehicle, "depth_{}.jpg".format(h))
            rgb_cam.save_image(env.vehicle, "rgb_{}.jpg".format(h))
        # env.engine.screenshot()

    env = SafeMetaDriveEnv(
        {
            "num_scenarios": 1,
            "traffic_density": 0.,
            "accident_prob": 1.,
            "start_seed": 4,
            "map": "SSSSS",
            "manual_control": True,
            "use_render": True,
            "image_observation": True,
            "rgb_clip": True,
            "interface_panel": ["depth_camera"],
            "sensors": dict(depth_camera=(DepthCamera, 800, 600)),
            "vehicle_config": dict(image_source="depth_camera"),
            # "map_config": {
            #     BaseMap.GENERATE_TYPE: MapGenerateMethod.BIG_BLOCK_NUM,
            #     BaseMap.GENERATE_CONFIG: 12,
            #     BaseMap.LANE_WIDTH: 3.5,
            #     BaseMap.LANE_NUM: 3,
            # }
        }
    )
    env.reset()
    env.engine.accept("m", get_image, extraArgs=[env])

    for i in range(1, 100000):
        o, r, tm, tc, info = env.step([0, 1])
        assert env.observation_space.contains(o)
        if env.config["use_render"]:
            # for i in range(ImageObservation.STACK_SIZE):
            #     ObservationType.show_gray_scale_array(o["image"][:, :, i])
            env.render()
        # if tm or tc:
        #     # print("Reset")
        #     env.reset()
    env.close()



"""

RenderPipeline

Copyright (c) 2014-2016 tobspr <tobias.springer1@gmail.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

"""

# This tool offers an interface to configure the pipeline

from __future__ import print_function

import os
import sys
import time
from threading import Thread
from functools import partial

# Change to the current directory
os.chdir(os.path.join(os.path.dirname(os.path.realpath(__file__))))

# Append the current directory to the path
sys.path.insert(0, os.getcwd())

# Add the render pipeline to the path
sys.path.insert(0, "../../")

from rplibs.six import iteritems  # noqa
from rplibs.pyqt_imports import *  #noqa

# Load the generated UI Layout
from ui.main_window_generated import Ui_MainWindow  # noqa

from rpcore.pluginbase.manager import PluginManager  # noqa
from rpcore.util.network_communication import NetworkCommunication  # noqa
from rpcore.mount_manager import MountManager  # noqa


class PluginConfigurator(QMainWindow, Ui_MainWindow):
    """ Interface to change the plugin settings """
    def __init__(self):

        # Init mounts
        self._mount_mgr = MountManager(None)
        self._mount_mgr.mount()

        self._plugin_mgr = PluginManager(None)
        self._plugin_mgr.requires_daytime_settings = False

        QMainWindow.__init__(self)
        Ui_MainWindow.__init__(self)
        self.setupUi(self)

        self._current_plugin = None
        self._current_plugin_instance = None
        self.lbl_restart_pipeline.hide()
        self._set_settings_visible(False)
        self._update_queue = list()

        qt_connect(self.lst_plugins, "itemSelectionChanged()", self.on_plugin_selected)
        qt_connect(self.lst_plugins, "itemChanged(QListWidgetItem*)", self.on_plugin_state_changed)
        qt_connect(self.btn_reset_plugin_settings, "clicked()", self.on_reset_plugin_settings)

        self._load_plugin_list()

        # Adjust column widths
        self.table_plugin_settings.setColumnWidth(0, 140)
        self.table_plugin_settings.setColumnWidth(1, 105)
        self.table_plugin_settings.setColumnWidth(2, 160)

        update_thread = Thread(target=self.update_thread, args=())
        update_thread.start()

    def closeEvent(self, event):  # noqa
        event.accept()
        import os
        os._exit(1)

    def on_reset_plugin_settings(self):
        """ Gets called when the user wants to reset settings of a plugin """

        # Ask the user if he's really sure about it
        msg = "Are you sure you want to reset the settings of '"
        msg += self._current_plugin_instance.name + "'?\n"
        msg += "This does not reset the Time of Day settings of this plugin.\n\n"
        msg += "!! This cannot be undone !! They will be lost forever (a long time!)."
        reply = QMessageBox.question(self, "Warning", msg, QMessageBox.Yes, QMessageBox.No)
        if reply == QMessageBox.Yes:

            QMessageBox.information(self, "Success", "Settings have been reset! You may have to restart the pipeline.")
            self._plugin_mgr.reset_plugin_settings(self._current_plugin)

            # Save config
            self._plugin_mgr.save_overrides("/$$rpconfig/plugins.yaml")

            # Always show the restart hint, even if its not always required
            self._show_restart_hint()

            # Re-render everything
            self._load_plugin_list()

    def on_plugin_state_changed(self, item):
        """ Handler when a plugin got activated/deactivated """
        plugin_id = item._plugin_id
        state = item.checkState() == Qt.Checked
        self._plugin_mgr.set_plugin_enabled(plugin_id, state)
        self._rewrite_plugin_config()
        self._show_restart_hint()

    def on_plugin_selected(self):
        """ Gets called when a plugin got selected in the plugin list """
        selected_item = self.lst_plugins.selectedItems()
        if not selected_item:
            self._current_plugin = None
            self._current_plugin_instance = None
            self._set_settings_visible(False)
            return
        assert len(selected_item) == 1
        selected_item = selected_item[0]
        self._current_plugin = selected_item._plugin_id
        self._current_plugin_instance = self._plugin_mgr.instances[self._current_plugin]
        assert (self._current_plugin_instance is not None)
        self._render_current_plugin()
        self._set_settings_visible(True)

    def update_thread(self):
        """ Internal update thread """
        while True:
            if len(self._update_queue) > 0:
                item = self._update_queue.pop(-1)
                NetworkCommunication.send_async(NetworkCommunication.CONFIG_PORT, item)

                if item.startswith("setval "):
                    setting_id = item.split()[1]
                    for entry in list(self._update_queue):
                        if entry.split()[1] == setting_id:
                            self._update_queue.remove(entry)
            time.sleep(0.2)

    def _rewrite_plugin_config(self):
        """ Rewrites the plugin configuration """
        self._plugin_mgr.save_overrides("/$$rpconfig/plugins.yaml")

    def _render_current_plugin(self):
        """ Displays the currently selected plugin """
        self.lbl_plugin_name.setText(
            self._current_plugin_instance.name.upper() + " <span style='color: #999; margin-left: 5px;'>[" +
            self._current_plugin_instance.plugin_id + "]</span>"
        )

        version_str = "Version " + self._current_plugin_instance.version
        version_str += " by " + self._current_plugin_instance.author

        self.lbl_plugin_version.setText(version_str)
        self.lbl_plugin_desc.setText(self._current_plugin_instance.description)

        self._render_current_settings()

    def _show_restart_hint(self):
        """ Shows a hint to restart the pipeline """
        self.lbl_restart_pipeline.show()

    def _render_current_settings(self):
        """ Renders the current plugin settings """
        settings = self._plugin_mgr.settings[self._current_plugin]

        # remove all rows
        while self.table_plugin_settings.rowCount() > 0:
            self.table_plugin_settings.removeRow(0)

        label_font = QFont()
        label_font.setPointSize(10)
        label_font.setFamily("Roboto")

        desc_font = QFont()
        desc_font.setPointSize(8)
        desc_font.setFamily("Roboto")

        for index, (name, handle) in enumerate(iteritems(settings)):
            if not handle.should_be_visible(settings):
                continue

            row_index = self.table_plugin_settings.rowCount()

            # Increase row count
            self.table_plugin_settings.insertRow(self.table_plugin_settings.rowCount())

            label = QLabel()
            label.setText(handle.label)
            label.setWordWrap(True)
            label.setFont(label_font)

            if not (handle.shader_runtime or handle.runtime):
                label.setStyleSheet("color: #999;")

            if handle.display_conditions:
                label.setStyleSheet(label.styleSheet() + "padding-left: 10px;")

            label.setMargin(10)

            self.table_plugin_settings.setCellWidget(row_index, 0, label)

            item_default = QTableWidgetItem()
            item_default.setText(str(handle.default))
            item_default.setTextAlignment(Qt.AlignCenter)
            self.table_plugin_settings.setItem(row_index, 1, item_default)

            setting_widget = self._get_widget_for_setting(name, handle)
            self.table_plugin_settings.setCellWidget(row_index, 2, setting_widget)

            label_desc = QLabel()
            label_desc.setText(handle.description)
            label_desc.setWordWrap(True)
            label_desc.setFont(desc_font)
            label_desc.setStyleSheet("color: #555;padding: 5px;")

            self.table_plugin_settings.setCellWidget(row_index, 3, label_desc)

    def _do_update_setting(self, setting_id, value):
        """ Updates a setting of the current plugin """

        # Check whether the setting is a runtime setting
        setting_handle = self._plugin_mgr.get_setting_handle(self._current_plugin, setting_id)

        # Skip the setting in case the value is equal
        if setting_handle.value == value:
            return

        # Otherwise set the new value
        setting_handle.set_value(value)
        self._rewrite_plugin_config()

        if not setting_handle.runtime and not setting_handle.shader_runtime:
            self._show_restart_hint()
        else:
            # In case the setting is dynamic, notice the pipeline about it:
            # print("Sending reload packet ...")
            self._update_queue.append("setval {}.{} {}".format(self._current_plugin, setting_id, value))

        # Update GUI, but only in case of enum and bool values, since they can trigger
        # display conditions:
        if setting_handle.type == "enum" or setting_handle.type == "bool":
            self._render_current_settings()

    def _on_setting_bool_changed(self, setting_id, value):
        self._do_update_setting(setting_id, value == Qt.Checked)

    def _on_setting_scalar_changed(self, setting_id, value):
        self._do_update_setting(setting_id, value)

    def _on_setting_enum_changed(self, setting_id, value):
        self._do_update_setting(setting_id, value)

    def _on_setting_power_of_two_changed(self, setting_id, value):
        self._do_update_setting(setting_id, value)

    def _on_setting_slider_changed(self, setting_id, setting_type, bound_objs, value):
        if setting_type == "float":
            value /= 100000.0
        self._do_update_setting(setting_id, value)

        for obj in bound_objs:
            obj.setValue(value)

    def _on_setting_spinbox_changed(self, setting_id, setting_type, bound_objs, value):
        self._do_update_setting(setting_id, value)
        # Assume objects are sliders, so we need to rescale the value
        for obj in bound_objs:
            obj.setValue(value * 100000.0 if setting_type == "float" else value)

    def _choose_path(self, setting_id, setting_handle, bound_objs):
        """ Shows a file chooser to show an path from """

        this_dir = os.path.dirname(os.path.realpath(__file__))
        plugin_dir = os.path.join(this_dir, "../../rpplugins/" + self._current_plugin, "resources")
        plugin_dir = os.path.abspath(plugin_dir)
        search_dir = os.path.join(plugin_dir, setting_handle.base_path)

        print("Plugin dir =", plugin_dir)
        print("Search dir =", search_dir)

        current_file = setting_handle.value.replace("\\", "/").split("/")[-1]
        print("Current file =", current_file)
        file_dlg = QFileDialog(self, "Choose File ..", search_dir, setting_handle.file_type)
        file_dlg.selectFile(current_file)
        # file_dlg.setViewMode(QFileDialog.Detail)

        if file_dlg.exec_():
            filename = file_dlg.selectedFiles()
            filename = filename[-1]
            print("QT selected files returned:", filename)

            filename = os.path.relpath(str(filename), plugin_dir)
            filename = filename.replace("\\", "/")
            print("Relative path is", filename)
            self._do_update_setting(setting_id, filename)

            display_file = filename.split("/")[-1]
            for obj in bound_objs:
                obj.setText(display_file)

    def _get_widget_for_setting(self, setting_id, setting):
        """ Returns an appropriate widget to control the given setting """

        widget = QWidget()
        layout = QHBoxLayout()
        layout.setAlignment(Qt.AlignCenter)
        widget.setLayout(layout)

        if setting.type == "bool":
            box = QCheckBox()
            box.setChecked(Qt.Checked if setting.value else Qt.Unchecked)
            qt_connect(box, "stateChanged(int)", partial(self._on_setting_bool_changed, setting_id))
            layout.addWidget(box)

        elif setting.type == "float" or setting.type == "int":

            if setting.type == "float":
                box = QDoubleSpinBox()

                if setting.maxval - setting.minval <= 2.0:
                    box.setDecimals(4)
            else:
                box = QSpinBox()
            box.setMinimum(setting.minval)
            box.setMaximum(setting.maxval)
            box.setValue(setting.value)

            box.setAlignment(Qt.AlignCenter)

            slider = QSlider()
            slider.setOrientation(Qt.Horizontal)

            if setting.type == "float":
                box.setSingleStep(abs(setting.maxval - setting.minval) / 100.0)
                slider.setMinimum(setting.minval * 100000.0)
                slider.setMaximum(setting.maxval * 100000.0)
                slider.setValue(setting.value * 100000.0)
            elif setting.type == "int":
                box.setSingleStep(max(1, (setting.maxval - setting.minval) / 32))
                slider.setMinimum(setting.minval)
                slider.setMaximum(setting.maxval)
                slider.setValue(setting.value)

            layout.addWidget(box)
            layout.addWidget(slider)

            qt_connect(
                slider, "valueChanged(int)", partial(self._on_setting_slider_changed, setting_id, setting.type, [box])
            )

            value_type = "int" if setting.type == "int" else "double"

            qt_connect(
                box, "valueChanged(" + value_type + ")",
                partial(self._on_setting_spinbox_changed, setting_id, setting.type, [slider])
            )

        elif setting.type == "enum":
            box = QComboBox()
            for value in setting.values:
                box.addItem(value)
            qt_connect(box, "currentIndexChanged(QString)", partial(self._on_setting_enum_changed, setting_id))
            box.setCurrentIndex(setting.values.index(setting.value))
            box.setMinimumWidth(145)

            layout.addWidget(box)

        elif setting.type == "power_of_two":

            box = QComboBox()
            resolutions = [str(2**i) for i in range(1, 32) if 2**i >= setting.minval and 2**i <= setting.maxval]
            for value in resolutions:
                box.addItem(value)
            qt_connect(box, "currentIndexChanged(QString)", partial(self._on_setting_power_of_two_changed, setting_id))
            box.setCurrentIndex(resolutions.index(str(setting.value)))
            box.setMinimumWidth(145)
            layout.addWidget(box)

        elif setting.type == "sample_sequence":

            box = QComboBox()
            for value in setting.sequences:
                box.addItem(value)
            qt_connect(box, "currentIndexChanged(QString)", partial(self._on_setting_enum_changed, setting_id))
            box.setCurrentIndex(setting.sequences.index(str(setting.value)))
            box.setMinimumWidth(145)
            layout.addWidget(box)

        elif setting.type == "path":

            label = QLabel()
            display_file = setting.value.replace("\\", "/").split("/")[-1]

            desc_font = QFont()
            desc_font.setPointSize(7)
            desc_font.setFamily("Roboto")

            label.setText(display_file)
            label.setFont(desc_font)

            button = QPushButton()
            button.setText("Choose File ...")
            qt_connect(button, "clicked()", partial(self._choose_path, setting_id, setting, (label, )))

            layout.addWidget(label)
            layout.addWidget(button)

        else:
            print("ERROR: Unkown setting type:", setting.type)

        return widget

    def _set_settings_visible(self, flag):
        """ Sets whether the settings panel is visible or not """
        if flag:
            self.frame_details.show()
        else:
            self.frame_details.hide()

    def _load_plugin_list(self):
        """ Reloads the whole plugin list """
        print("Loading plugin list")

        # Reset selection
        self._current_plugin = None
        self._current_plugin_instance = None
        self._set_settings_visible(False)

        # Plugins are all plugins in the plugins directory
        self._plugin_mgr.unload()
        self._plugin_mgr.load()

        self.lst_plugins.clear()
        plugins = sorted(iteritems(self._plugin_mgr.instances), key=lambda plg: plg[1].name)

        item_font = QFont()
        item_font.setBold(False)
        item_font.setPointSize(10)

        for plugin_id, instance in plugins:

            item = QListWidgetItem()
            item.setText(" " + instance.name)
            item.setFont(item_font)

            if self._plugin_mgr.is_plugin_enabled(plugin_id):
                item.setCheckState(Qt.Checked)
            else:
                item.setCheckState(Qt.Unchecked)

            item._plugin_id = plugin_id
            self.lst_plugins.addItem(item)

        self.lst_plugins.setCurrentRow(0)


# Start application
app = QApplication(sys.argv)
qt_register_fonts()
configurator = PluginConfigurator()
configurator.show()
app.exec_()



from metadrive.envs.safe_metadrive_env import SafeMetaDriveEnv


def test_episode_release():
    try:
        env = SafeMetaDriveEnv(
            {
                "use_render": False,
                "num_scenarios": 100,
                "accident_prob": .8,
                "traffic_density": 0.5,
                "debug": True
            }
        )
        o, _ = env.reset()
        for i in range(1, 10):
            env.step([1.0, 1.0])
            env.step([1.0, 1.0])
            env.step([1.0, 1.0])
            env.step([1.0, 1.0])
            env.reset()
            env.reset()
    finally:
        env.close()



"""

RenderTarget

Copyright (c) 2015 tobspr <tobias.springer1@gmail.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

"""

from __future__ import print_function, division

from panda3d.core import GraphicsOutput, Texture, AuxBitplaneAttrib, NodePath
from panda3d.core import Vec4, TransparencyAttrib, ColorWriteAttrib, SamplerState
from panda3d.core import WindowProperties, FrameBufferProperties, GraphicsPipe
from panda3d.core import LVecBase2i

from metadrive.render_pipeline.rplibs.six.moves import range  # pylint: disable=import-error
from metadrive.render_pipeline.rplibs.six import iterkeys, itervalues

from metadrive.render_pipeline.rpcore.globals import Globals
from metadrive.render_pipeline.rpcore.rpobject import RPObject
from metadrive.render_pipeline.rpcore.util.post_process_region import PostProcessRegion

__all__ = "RenderTarget",
__version__ = "2.0"


class setter(object):  # noqa # pylint: disable=invalid-name,too-few-public-methods
    """ Setter only property """
    def __init__(self, func):
        self.__func = func
        self.__doc__ = func.__doc__

    def __set__(self, name, value):
        return self.__func(name, value)


class RenderTarget(RPObject):
    """ Second version of the RenderTarget library, provides functions
    to easily setup buffers in Panda3D. """

    NUM_ALLOCATED_BUFFERS = 0
    USE_R11G11B10 = True
    REGISTERED_TARGETS = []
    CURRENT_SORT = -300

    def __init__(self, name="target"):
        RPObject.__init__(self, name)
        self._targets = {}
        self._color_bits = (0, 0, 0, 0)
        self._aux_bits = 8
        self._aux_count = 0
        self._depth_bits = 0
        self._size = LVecBase2i(-1)
        self._size_constraint = LVecBase2i(-1)
        self._source_window = Globals.base.win
        self._source_region = None
        self._active = False
        self._internal_buffer = None
        self.sort = None

        # Public attributes
        self.engine = Globals.base.graphicsEngine
        self.support_transparency = False
        self.create_default_region = True

        # Disable all global clears, since they are not required
        for region in Globals.base.win.get_display_regions():
            region.disable_clears()

    def add_color_attachment(self, bits=8, alpha=False):
        """ Adds a new color attachment with the given amount of bits, bits can
        be either a single int or a tuple determining the bits. If bits is a
        single int, alpha determines whether alpha bits are requested """
        self._targets["color"] = Texture(self.debug_name + "_color")
        if isinstance(bits, (list, tuple)):
            self._color_bits = (bits[0], bits[1], bits[2], bits[3] if len(bits) == 4 else 0)
        else:
            self._color_bits = ((bits, bits, bits, (bits if alpha else 0)))

    def add_depth_attachment(self, bits=32):
        """ Adds a depth attachment wit the given amount of bits """
        self._targets["depth"] = Texture(self.debug_name + "_depth")
        self._depth_bits = bits

    def add_aux_attachment(self, bits=8):
        """ Adds a new aux attachment with the given amount of bits. The amount
        of bits passed overrides all previous bits set, since all aux textures
        have to have the same amount of bits. """
        self._aux_bits = bits
        self._aux_count += 1

    def add_aux_attachments(self, bits=8, count=1):
        """ Adds n new aux attachments, with the given amount of bits. All
        previously set aux bits are overriden, since all aux textures have to
        have the same amount of bits """
        self._aux_bits = bits
        self._aux_count += count

    @setter
    def size(self, *args):
        """ Sets the render target size. This can be either a single integer,
        in which case it applies to both dimensions. Negative integers cause
        the render target to be proportional to the screen size, i.e. a value
        of -4 produces a quarter resolution target, a value of -2 a half
        resolution target, and a value of -1 a full resolution target
        (the default). """
        self._size_constraint = LVecBase2i(*args)

    @property
    def active(self):
        """ Returns whether the target is currently active """
        return self._active

    @active.setter
    def active(self, flag):
        """ Sets whether the target is active, this just propagates the active
        flag to all display regions """
        for region in self._internal_buffer.get_display_regions():
            region.set_active(flag)

    @property
    def color_tex(self):
        """ Returns the color attachment if present """
        return self._targets["color"]

    @property
    def depth_tex(self):
        """ Returns the depth attachment if present """
        return self._targets["depth"]

    @property
    def aux_tex(self):
        """ Returns a list of aux textures, can be used like target.aux_tex[2],
        notice the indices start at zero, so the first target has the index 0. """
        return [self._targets[i] for i in sorted(iterkeys(self._targets)) if i.startswith("aux_")]

    def set_shader_input(self, *args, **kwargs):
        """ Sets a shader input available to the target """
        if self.create_default_region:
            self._source_region.set_shader_input(*args, **kwargs)

    def set_shader_inputs(self, **kwargs):
        """ Sets shader inputs available to the target """
        if self.create_default_region:
            self._source_region.set_shader_inputs(**kwargs)

    @setter
    def shader(self, shader_obj):
        """ Sets a shader on the target """
        if not shader_obj:
            self.error("shader must not be None!")
            return
        self._source_region.set_shader(shader_obj)

    @property
    def internal_buffer(self):
        """ Returns a handle to the internal GraphicsBuffer object """
        return self._internal_buffer

    @property
    def targets(self):
        """ Returns the dictionary of attachments, whereas the key is the name
        of the attachment and the value is the Texture handle of the attachment """
        return self._targets

    @property
    def region(self):
        """ Returns the internally used PostProcessRegion """
        return self._source_region

    def prepare_render(self, camera_np):
        """ Prepares to render a scene """
        self.create_default_region = False
        self._create_buffer()
        self._source_region = self._internal_buffer.get_display_region(0)

        if camera_np:
            initial_state = NodePath("rtis")
            initial_state.set_state(camera_np.node().get_initial_state())

            if self._aux_count:
                initial_state.set_attrib(AuxBitplaneAttrib.make(self._aux_bits), 20)
            initial_state.set_attrib(TransparencyAttrib.make(TransparencyAttrib.M_none), 20)

            if max(self._color_bits) == 0:
                initial_state.set_attrib(ColorWriteAttrib.make(ColorWriteAttrib.C_off), 20)

            # Disable existing regions of the camera
            for region in camera_np.node().get_display_regions():
                region.set_active(False)

            # Remove the existing display region of the camera
            for region in self._source_window.get_display_regions():
                if region.get_camera() == camera_np:
                    self._source_window.remove_display_region(region)

            camera_np.node().set_initial_state(initial_state.get_state())
            self._source_region.set_camera(camera_np)

        self._internal_buffer.disable_clears()
        self._source_region.disable_clears()
        self._source_region.set_active(True)
        self._source_region.set_sort(20)

        # Reenable depth-clear, usually desireable
        self._source_region.set_clear_depth_active(True)
        self._source_region.set_clear_depth(1.0)
        self._active = True

    def prepare_buffer(self):
        """ Prepares the target to render to an offscreen buffer """
        self._create_buffer()
        self._active = True

    def present_on_screen(self):
        """ Prepares the target to render on the main window, to present the
        final rendered image """
        self._source_region = PostProcessRegion.make(self._source_window)
        self._source_region.set_sort(5)

    def remove(self):
        """ Deletes this buffer, restoring the previous state """
        self._internal_buffer.clear_render_textures()
        self.engine.remove_window(self._internal_buffer)
        self._active = False
        for target in itervalues(self._targets):
            target.release_all()
        RenderTarget.REGISTERED_TARGETS.remove(self)

    def set_clear_color(self, *args):
        """ Sets the  clear color """
        self._internal_buffer.set_clear_color_active(True)
        self._internal_buffer.set_clear_color(Vec4(*args))

    @setter
    def instance_count(self, count):
        """ Sets the instance count """
        self._source_region.set_instance_count(count)

    def _create_buffer(self):
        """ Internal method to create the buffer object """
        self._compute_size_from_constraint()
        if not self._create():
            self.error("Failed to create buffer!")
            return False

        if self.create_default_region:
            self._source_region = PostProcessRegion.make(self._internal_buffer)

            if max(self._color_bits) == 0:
                self._source_region.set_attrib(ColorWriteAttrib.make(ColorWriteAttrib.M_none), 1000)

    def _compute_size_from_constraint(self):
        """ Computes the actual size in pixels from the targets size constraint """
        w, h = Globals.resolution.x, Globals.resolution.y
        self._size = LVecBase2i(self._size_constraint)
        if self._size_constraint.x < 0:
            self._size.x = (w - self._size_constraint.x - 1) // (-self._size_constraint.x)
        if self._size_constraint.y < 0:
            self._size.y = (h - self._size_constraint.y - 1) // (-self._size_constraint.y)

    def _setup_textures(self):
        """ Prepares all bound textures """
        for i in range(self._aux_count):
            self._targets["aux_{}".format(i)] = Texture(self.debug_name + "_aux{}".format(i))
        for tex in itervalues(self._targets):
            tex.set_wrap_u(SamplerState.WM_clamp)
            tex.set_wrap_v(SamplerState.WM_clamp)
            tex.set_anisotropic_degree(0)
            tex.set_x_size(self._size.x)
            tex.set_y_size(self._size.y)
            tex.set_minfilter(SamplerState.FT_linear)
            tex.set_magfilter(SamplerState.FT_linear)

    def _make_properties(self):
        """ Creates the window and buffer properties """
        window_props = WindowProperties.size(self._size.x, self._size.y)
        buffer_props = FrameBufferProperties()

        if self._size_constraint.x == 0 or self._size_constraint.y == 0:
            window_props = WindowProperties.size(1, 1)

        if self._color_bits == (16, 16, 16, 0):
            if RenderTarget.USE_R11G11B10:
                buffer_props.set_rgba_bits(11, 11, 10, 0)
            else:
                buffer_props.set_rgba_bits(*self._color_bits)
        elif 8 in self._color_bits:
            # When specifying 8 bits, specify 1 bit, this is a workarround
            # to a legacy logic in panda
            buffer_props.set_rgba_bits(*[i if i != 8 else 1 for i in self._color_bits])
        else:
            buffer_props.set_rgba_bits(*self._color_bits)

        buffer_props.set_accum_bits(0)
        buffer_props.set_stencil_bits(0)
        buffer_props.set_back_buffers(0)
        buffer_props.set_coverage_samples(0)
        buffer_props.set_depth_bits(self._depth_bits)

        if self._depth_bits == 32:
            buffer_props.set_float_depth(True)

        buffer_props.set_float_color(max(self._color_bits) > 8)

        buffer_props.set_force_hardware(True)
        buffer_props.set_multisamples(0)
        buffer_props.set_srgb_color(False)
        buffer_props.set_stereo(False)
        buffer_props.set_stencil_bits(0)

        if self._aux_bits == 8:
            buffer_props.set_aux_rgba(self._aux_count)
        elif self._aux_bits == 16:
            buffer_props.set_aux_hrgba(self._aux_count)
        elif self._aux_bits == 32:
            buffer_props.set_aux_float(self._aux_count)
        else:
            self.error("Invalid aux bits")

        return window_props, buffer_props

    def _create(self):
        """ Creates the internally used buffer """
        self._setup_textures()
        window_props, buffer_props = self._make_properties()

        self._internal_buffer = self.engine.make_output(
            self._source_window.get_pipe(), self.debug_name, 1, buffer_props, window_props,
            GraphicsPipe.BF_refuse_window | GraphicsPipe.BF_resizeable, self._source_window.gsg, self._source_window
        )

        if not self._internal_buffer:
            self.error("Failed to create buffer")
            return

        if self._depth_bits:
            self._internal_buffer.add_render_texture(
                self.depth_tex, GraphicsOutput.RTM_bind_or_copy, GraphicsOutput.RTP_depth
            )

        if max(self._color_bits) > 0:
            self._internal_buffer.add_render_texture(
                self.color_tex, GraphicsOutput.RTM_bind_or_copy, GraphicsOutput.RTP_color
            )

        aux_prefix = {
            8: "RTP_aux_rgba_{}",
            16: "RTP_aux_hrgba_{}",
            32: "RTP_aux_float_{}",
        }[self._aux_bits]

        for i in range(self._aux_count):
            target_mode = getattr(GraphicsOutput, aux_prefix.format(i))
            self._internal_buffer.add_render_texture(self.aux_tex[i], GraphicsOutput.RTM_bind_or_copy, target_mode)

        if not self.sort:
            RenderTarget.CURRENT_SORT += 20
            self.sort = RenderTarget.CURRENT_SORT

        RenderTarget.NUM_ALLOCATED_BUFFERS += 1
        self._internal_buffer.set_sort(self.sort)
        self._internal_buffer.disable_clears()
        self._internal_buffer.get_display_region(0).disable_clears()
        self._internal_buffer.get_overlay_display_region().disable_clears()
        self._internal_buffer.get_overlay_display_region().set_active(False)

        RenderTarget.REGISTERED_TARGETS.append(self)
        return True

    def consider_resize(self):
        """ Checks if the target has to get resized, and if this is the case,
        performs the resize. This should be called when the window resolution
        changed. """
        current_size = LVecBase2i(self._size)
        self._compute_size_from_constraint()
        if current_size != self._size:
            if self._internal_buffer:
                self._internal_buffer.set_size(self._size.x, self._size.y)



"""

RenderPipeline

Copyright (c) 2014-2016 tobspr <tobias.springer1@gmail.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

"""

from __future__ import print_function

from metadrive.render_pipeline.rplibs.six.moves import range  # pylint: disable=import-error

from panda3d.core import Camera, MatrixLens

from metadrive.render_pipeline.rpcore.pynative.shadow_atlas import ShadowAtlas


class ShadowManager(object):
    """ Please refer to the native C++ implementation for docstrings and comments.
    This is just the python implementation, which does not contain documentation! """
    def __init__(self):
        self._max_updates = 10
        self._atlas = None
        self._atlas_size = 4096
        self._tag_state_mgr = None
        self._atlas_graphics_output = None
        self._display_regions = []
        self._queued_updates = []
        self._cameras = []
        self._camera_nps = []

    def set_max_updates(self, max_updates):
        if max_updates == 0:
            print("Warning: max_updates set to 0, no shadow updates will happen")
        self._max_updates = max_updates

    def set_atlas_size(self, atlas_size):
        self._atlas_size = atlas_size

    def get_atlas_size(self):
        return self._atlas_size

    atlas_size = property(get_atlas_size, set_atlas_size)

    def set_scene(self, scene_parent):
        self._scene_parent = scene_parent

    def set_tag_state_manager(self, tag_mgr):
        self._tag_state_mgr = tag_mgr

    def set_atlas_graphics_output(self, graphics_output):
        self._atlas_graphics_output = graphics_output

    def get_num_update_slots_left(self):
        return self._max_updates - len(self._queued_updates)

    num_update_slots_left = property(get_num_update_slots_left)

    def get_atlas(self):
        return self._atlas

    atlas = property(get_atlas)

    def init(self):
        for i in range(self._max_updates):
            camera = Camera("ShadowCam-" + str(i))
            camera.set_lens(MatrixLens())
            camera.set_active(False)
            camera.set_scene(self._scene_parent)
            self._tag_state_mgr.register_camera("shadow", camera)
            self._camera_nps.append(self._scene_parent.attach_new_node(camera))
            self._cameras.append(camera)

            region = self._atlas_graphics_output.make_display_region()
            region.set_sort(1000)
            region.set_clear_depth_active(True)
            region.set_clear_depth(1.0)
            region.set_clear_color_active(False)
            region.set_camera(self._camera_nps[i])
            region.set_active(False)
            self._display_regions.append(region)

        self._atlas = ShadowAtlas(self._atlas_size)

    def update(self):
        for i in range(len(self._queued_updates), self._max_updates):
            self._cameras[i].set_active(False)
            self._display_regions[i].set_active(False)

        for i, source in enumerate(self._queued_updates):
            self._cameras[i].set_active(True)
            self._display_regions[i].set_active(True)
            self._cameras[i].get_lens().set_user_mat(source.get_mvp())
            uv = source.get_uv_region()
            self._display_regions[i].set_dimensions(uv.x, uv.x + uv.z, uv.y, uv.y + uv.w)

        self._queued_updates = []

    def add_update(self, source):
        if len(self._queued_updates) >= self._max_updates:
            return False
        self._queued_updates.append(source)
        return True



import copy

import numpy as np

from metadrive.component.pgblock.create_pg_block_utils import CreateAdverseRoad, CreateRoadFrom, ExtendStraightLane, \
    CreateTwoWayRoad, create_bend_straight
from metadrive.component.pgblock.pg_block import PGBlock, PGBlockSocket
from metadrive.component.road_network import Road
from metadrive.constants import PGLineType, PGLineColor
from metadrive.component.pg_space import ParameterSpace, Parameter, BlockParameterSpace


class ParkingLot(PGBlock):
    """
    Parking Lot
    """

    ID = "P"
    PARAMETER_SPACE = ParameterSpace(BlockParameterSpace.PARKING_LOT_PARAMETER)
    ANGLE = np.deg2rad(90)
    SOCKET_LENGTH = 4  # m
    SOCKET_NUM = 1

    def _try_plug_into_previous_block(self) -> bool:
        self.spawn_roads = []
        self.dest_roads = []

        no_cross = True
        para = self.get_config()
        assert self.positive_lane_num == 1, "Lane number of previous block must be 1 in each direction"

        self.parking_space_length = para[Parameter.length]
        self.parking_space_width = self.lane_width
        parking_space_num = para[Parameter.one_side_vehicle_num]
        # parking_space_num = 10
        radius = para[Parameter.radius]

        main_straight_road_length = 2 * radius + (parking_space_num - 1) * self.parking_space_width
        main_lane = ExtendStraightLane(
            self.positive_lanes[0], main_straight_road_length, [PGLineType.BROKEN, PGLineType.NONE]
        )
        road = Road(self.pre_block_socket.positive_road.end_node, self.road_node(0, 0))

        # main straight part
        no_cross = CreateRoadFrom(
            main_lane,
            self.positive_lane_num,
            road,
            self.block_network,
            self._global_network,
            center_line_type=PGLineType.BROKEN,
            inner_lane_line_type=PGLineType.BROKEN,
            side_lane_line_type=PGLineType.NONE,
            center_line_color=PGLineColor.GREY,
            ignore_intersection_checking=self.ignore_intersection_checking
        ) and no_cross
        no_cross = CreateAdverseRoad(
            road,
            self.block_network,
            self._global_network,
            center_line_type=PGLineType.BROKEN,
            inner_lane_line_type=PGLineType.BROKEN,
            side_lane_line_type=PGLineType.NONE,
            center_line_color=PGLineColor.GREY,
            ignore_intersection_checking=self.ignore_intersection_checking
        ) and no_cross

        # socket part
        parking_lot_out_lane = ExtendStraightLane(main_lane, self.SOCKET_LENGTH, [PGLineType.BROKEN, PGLineType.NONE])
        parking_lot_out_road = Road(self.road_node(0, 0), self.road_node(0, 1))

        # out socket part
        no_cross = CreateRoadFrom(
            parking_lot_out_lane,
            self.positive_lane_num,
            parking_lot_out_road,
            self.block_network,
            self._global_network,
            center_line_type=PGLineType.BROKEN,
            inner_lane_line_type=PGLineType.BROKEN,
            side_lane_line_type=PGLineType.SIDE,
            ignore_intersection_checking=self.ignore_intersection_checking
        ) and no_cross

        no_cross = CreateAdverseRoad(
            parking_lot_out_road,
            self.block_network,
            self._global_network,
            center_line_type=PGLineType.BROKEN,
            inner_lane_line_type=PGLineType.BROKEN,
            side_lane_line_type=PGLineType.SIDE,
            ignore_intersection_checking=self.ignore_intersection_checking
        ) and no_cross

        socket = self.create_socket_from_positive_road(parking_lot_out_road)
        self.add_sockets(socket)

        # add parking space
        for i in range(int(parking_space_num)):
            no_cross = self._add_one_parking_space(
                copy.copy(self.get_socket_list()[0]).get_socket_in_reverse(),
                self.pre_block_socket.get_socket_in_reverse(), i + 1, radius, i * self.parking_space_width,
                (parking_space_num - i - 1) * self.parking_space_width
            ) and no_cross

        for i in range(parking_space_num, 2 * parking_space_num):
            index = i + 1
            i -= parking_space_num
            no_cross = self._add_one_parking_space(
                self.pre_block_socket, copy.copy(self.get_socket_list()[0]), index, radius,
                i * self.parking_space_width, (parking_space_num - i - 1) * self.parking_space_width
            ) and no_cross

        return no_cross

    def _add_one_parking_space(
        self, in_socket: PGBlockSocket, out_socket: PGBlockSocket, part_idx: int, radius, dist_to_in, dist_to_out
    ) -> bool:
        no_cross = True

        # lane into parking space and parking space, 1
        if in_socket.is_same_socket(self.pre_block_socket) or in_socket.is_same_socket(
                self.pre_block_socket.get_socket_in_reverse()):
            net = self._global_network
        else:
            net = self.block_network
        in_lane = in_socket.positive_road.get_lanes(net)[0]
        start_node = in_socket.positive_road.end_node
        if dist_to_in > 1e-3:
            # a straight part will be added
            in_lane = ExtendStraightLane(in_lane, dist_to_in, [PGLineType.NONE, PGLineType.NONE])
            in_road = Road(in_socket.positive_road.end_node, self.road_node(part_idx, 0))
            CreateRoadFrom(
                in_lane,
                self.positive_lane_num,
                in_road,
                self.block_network,
                self._global_network,
                center_line_type=PGLineType.NONE,
                inner_lane_line_type=PGLineType.NONE,
                side_lane_line_type=PGLineType.NONE,
                ignore_intersection_checking=self.ignore_intersection_checking
            )
            start_node = self.road_node(part_idx, 0)

        bend, straight = create_bend_straight(
            in_lane, self.parking_space_length, radius, self.ANGLE, True, self.parking_space_width
        )
        bend_road = Road(start_node, self.road_node(part_idx, 1))
        bend_no_cross = CreateRoadFrom(
            bend,
            self.positive_lane_num,
            bend_road,
            self.block_network,
            self._global_network,
            center_line_type=PGLineType.NONE,
            inner_lane_line_type=PGLineType.NONE,
            side_lane_line_type=PGLineType.SIDE if dist_to_in < 1e-3 else PGLineType.NONE,
            ignore_intersection_checking=self.ignore_intersection_checking
        )
        if dist_to_in < 1e-3:
            no_cross = no_cross and bend_no_cross

        straight_road = Road(self.road_node(part_idx, 1), self.road_node(part_idx, 2))
        self.dest_roads.append(straight_road)
        no_cross = no_cross and CreateRoadFrom(
            straight,
            self.positive_lane_num,
            straight_road,
            self.block_network,
            self._global_network,
            center_line_type=PGLineType.CONTINUOUS,
            inner_lane_line_type=PGLineType.NONE,
            side_lane_line_type=PGLineType.SIDE if dist_to_in < 1e-3 else PGLineType.NONE,
            center_line_color=PGLineColor.GREY,
            ignore_intersection_checking=self.ignore_intersection_checking
        )

        # lane into parking space and parking space, 2
        neg_road: Road = out_socket.negative_road
        if out_socket.is_same_socket(self.pre_block_socket) or out_socket.is_same_socket(
                self.pre_block_socket.get_socket_in_reverse()):
            net = self._global_network
        else:
            net = self.block_network
        neg_lane = \
            neg_road.get_lanes(net)[0]
        start_node = neg_road.end_node
        if dist_to_out > 1e-3:
            # a straight part will be added
            neg_lane = ExtendStraightLane(neg_lane, dist_to_out, [PGLineType.NONE, PGLineType.NONE])
            neg_road = Road(neg_road.end_node, self.road_node(part_idx, 3))
            CreateRoadFrom(
                neg_lane,
                self.positive_lane_num,
                neg_road,
                self.block_network,
                self._global_network,
                center_line_type=PGLineType.NONE,
                inner_lane_line_type=PGLineType.NONE,
                side_lane_line_type=PGLineType.NONE,
                ignore_intersection_checking=self.ignore_intersection_checking
            )
            start_node = self.road_node(part_idx, 3)

        bend, straight = create_bend_straight(
            neg_lane, self.lane_width, radius, self.ANGLE, False, self.parking_space_width
        )
        bend_road = Road(start_node, self.road_node(part_idx, 4))
        CreateRoadFrom(
            bend,
            self.positive_lane_num,
            bend_road,
            self.block_network,
            self._global_network,
            center_line_type=PGLineType.NONE,
            inner_lane_line_type=PGLineType.NONE,
            side_lane_line_type=PGLineType.NONE,
            ignore_intersection_checking=self.ignore_intersection_checking
        )

        straight_road = Road(self.road_node(part_idx, 4), self.road_node(part_idx, 1))
        CreateRoadFrom(
            straight,
            self.positive_lane_num,
            straight_road,
            self.block_network,
            self._global_network,
            center_line_type=PGLineType.NONE,
            inner_lane_line_type=PGLineType.NONE,
            side_lane_line_type=PGLineType.NONE,
            ignore_intersection_checking=self.ignore_intersection_checking
        )

        # give it a new road name to make it be a two way road (1,2) = (5,6) = parking space !
        parking_road = Road(self.road_node(part_idx, 5), self.road_node(part_idx, 6))
        self.spawn_roads.append(parking_road)
        CreateTwoWayRoad(
            Road(self.road_node(part_idx, 1), self.road_node(part_idx, 2)),
            self.block_network,
            self._global_network,
            parking_road,
            center_line_type=PGLineType.NONE,
            inner_lane_line_type=PGLineType.NONE,
            side_lane_line_type=PGLineType.SIDE if dist_to_out < 1e-3 else PGLineType.NONE,
            ignore_intersection_checking=self.ignore_intersection_checking
        )

        # out part
        parking_lane = parking_road.get_lanes(self.block_network)[0]

        # out part 1
        bend, straight = create_bend_straight(
            parking_lane, 0.1 if dist_to_out < 1e-3 else dist_to_out, radius, self.ANGLE, True, parking_lane.width
        )
        out_bend_road = Road(
            self.road_node(part_idx, 6),
            self.road_node(part_idx, 7) if dist_to_out > 1e-3 else out_socket.positive_road.start_node
        )
        bend_success = CreateRoadFrom(
            bend,
            self.positive_lane_num,
            out_bend_road,
            self.block_network,
            self._global_network,
            center_line_type=PGLineType.NONE,
            inner_lane_line_type=PGLineType.NONE,
            side_lane_line_type=PGLineType.SIDE if dist_to_out < 1e-3 else PGLineType.NONE,
            ignore_intersection_checking=self.ignore_intersection_checking
        )
        if dist_to_out < 1e-3:
            no_cross = no_cross and bend_success

        if dist_to_out > 1e-3:
            out_straight_road = Road(self.road_node(part_idx, 7), out_socket.positive_road.start_node)
            no_cross = no_cross and CreateRoadFrom(
                straight,
                self.positive_lane_num,
                out_straight_road,
                self.block_network,
                self._global_network,
                center_line_type=PGLineType.NONE,
                inner_lane_line_type=PGLineType.NONE,
                side_lane_line_type=PGLineType.NONE,
                ignore_intersection_checking=self.ignore_intersection_checking
            )
        # out part 2
        extend_lane = ExtendStraightLane(parking_lane, self.lane_width, [PGLineType.NONE, PGLineType.NONE])
        CreateRoadFrom(
            extend_lane,
            self.positive_lane_num,
            Road(self.road_node(part_idx, 6), self.road_node(part_idx, 8)),
            self.block_network,
            self._global_network,
            center_line_type=PGLineType.NONE,
            inner_lane_line_type=PGLineType.NONE,
            side_lane_line_type=PGLineType.NONE,
            ignore_intersection_checking=self.ignore_intersection_checking
        )

        bend, straight = create_bend_straight(
            extend_lane, 0.1 if dist_to_in < 1e-3 else dist_to_in, radius, self.ANGLE, False, parking_lane.width
        )
        out_bend_road = Road(
            self.road_node(part_idx, 8),
            self.road_node(part_idx, 9) if dist_to_in > 1e-3 else in_socket.negative_road.start_node
        )
        CreateRoadFrom(
            bend,
            self.positive_lane_num,
            out_bend_road,
            self.block_network,
            self._global_network,
            center_line_type=PGLineType.NONE,
            inner_lane_line_type=PGLineType.NONE,
            side_lane_line_type=PGLineType.NONE,
            ignore_intersection_checking=self.ignore_intersection_checking
        )
        if dist_to_in > 1e-3:
            out_straight_road = Road(self.road_node(part_idx, 9), in_socket.negative_road.start_node)
            CreateRoadFrom(
                straight,
                self.positive_lane_num,
                out_straight_road,
                self.block_network,
                self._global_network,
                center_line_type=PGLineType.NONE,
                inner_lane_line_type=PGLineType.NONE,
                side_lane_line_type=PGLineType.NONE,
                ignore_intersection_checking=self.ignore_intersection_checking
            )

        return no_cross

    @staticmethod
    def in_direction_parking_space(road: Road):
        """
        Give a parking space in out-direction, return in direction road
        """
        start_node = copy.deepcopy(road.start_node)
        end_node = copy.deepcopy(road.end_node)
        assert start_node[-2] == "5" and end_node[
            -2] == "6", "It is not out-direction of this parking space, start_node:{}, end_node:{}".format(
                start_node, end_node
            )
        start_node = start_node[:-2] + "1" + PGBlock.DASH
        end_node = end_node[:-2] + "2" + PGBlock.DASH
        return Road(start_node, end_node)

    @staticmethod
    def out_direction_parking_space(road: Road):
        """
        Give a parking space in in-direction, return out-direction road
        """
        start_node = copy.deepcopy(road.start_node)
        end_node = copy.deepcopy(road.end_node)
        assert start_node[-2] == "1" and end_node[
            -2] == "2", "It is not in-direction of this parking space, start_node:{}, end_node:{}".format(
                start_node, end_node
            )
        start_node = start_node[:-2] + "5" + PGBlock.DASH
        end_node = end_node[:-2] + "6" + PGBlock.DASH
        return Road(start_node, end_node)

    @staticmethod
    def is_out_direction_parking_space(road: Road):
        start_node = road.start_node
        end_node = road.end_node
        assert (start_node[-2] == "1"
                and end_node[-2] == "2") or (start_node[-2] == "5" and end_node[-2]
                                             == "6"), "{} to {} is not parking space".format(start_node, end_node)
        if start_node[-2] == "5" and end_node[-2] == "6":
            return True
        else:
            return False

    @staticmethod
    def is_in_direction_parking_space(road: Road):
        start_node = road.start_node
        end_node = road.end_node
        assert (start_node[-2] == "1"
                and end_node[-2] == "2") or (start_node[-2] == "5" and end_node[-2]
                                             == "6"), "{} to {} is not parking space".format(start_node, end_node)
        if start_node[-2] == "1" and end_node[-2] == "2":
            return True
        else:
            return False



from metadrive.envs.metadrive_env import MetaDriveEnv


def local_test_apply_action():
    try:
        env = MetaDriveEnv({"map": "SSS", "use_render": True})
        o, _ = env.reset()
        for act in [-1, 1]:
            for _ in range(300):
                assert env.observation_space.contains(o)
                o, r, tm, tc, i = env.step([act, 1])
                if tm or tc:
                    o, _ = env.reset()
                    break
        env.close()
    finally:
        if "env" in locals():
            env = locals()["env"]
            env.close()


if __name__ == '__main__':
    local_test_apply_action()



"""
Please feel free to run this script to enjoy a journey by keyboard!
Remember to press H to see help message!

Note: This script require rendering, please following the installation instruction to setup a proper
environment that allows popping up an window.
"""
import argparse
import logging
import random

import cv2
import numpy as np

from metadrive import MetaDriveEnv
from metadrive.component.sensors.rgb_camera import RGBCamera
from metadrive.constants import HELP_MESSAGE

if __name__ == "__main__":
    config = dict(
        # controller="joystick",
        use_render=True,
        manual_control=True,
        traffic_density=0.1,
        num_scenarios=10000,
        random_agent_model=False,
        random_lane_width=True,
        random_lane_num=True,
        vehicle_config=dict(show_lidar=False, show_navi_mark=False),
        # debug=True,
        # debug_static_world=True,
        map=4,  # seven block
        start_seed=10,
    )
    parser = argparse.ArgumentParser()
    parser.add_argument("--observation", type=str, default="lidar", choices=["lidar", "rgb_camera"])
    args = parser.parse_args()
    if args.observation == "rgb_camera":
        config.update(
            dict(
                image_observation=True,
                sensors=dict(rgb_camera=(RGBCamera, 512, 256)),
                interface_panel=["rgb_camera", "dashboard"]
            )
        )
    else:
        config["vehicle_config"]["show_lidar"] = True
    env = MetaDriveEnv(config)
    try:
        o, _ = env.reset(seed=21)
        print(HELP_MESSAGE)
        env.vehicle.expert_takeover = True
        if args.observation == "rgb_camera":
            assert isinstance(o, dict)
            print("The observation is a dict with numpy arrays as values: ", {k: v.shape for k, v in o.items()})
        else:
            assert isinstance(o, np.ndarray)
            print("The observation is an numpy array with shape: ", o.shape)
        for i in range(1, 1000000000):
            o, r, tm, tc, info = env.step([0, 0])
            env.render(
                text={
                    "Auto-Drive (Switch mode: T)": "on" if env.current_track_vehicle.expert_takeover else "off",
                    "Current Observation": args.observation,
                    "Keyboard Control": "W,A,S,D",
                }
            )

            if args.observation == "rgb_camera":
                cv2.imshow('RGB Image in Observation', o["image"][..., -1])
                cv2.waitKey(1)
            if (tm or tc) and info["arrive_dest"]:
                env.reset(env.current_seed + 1)
                env.current_track_vehicle.expert_takeover = True
    except Exception as e:
        raise e
    finally:
        env.close()
